"""
Very inefficient implementation of Pade approximation,
described in H. J. Vidberg and J. W. Serene, J. Low Temp. Phys. 29, 179 (1977).

Author: Josef Kaufmann
"""

import numpy as np
import sys


def compute_coefficients(zi,ui):
  cdef int i,j,n
  n=len(zi)
  g=np.zeros((n,n),dtype=np.complex)
  g[:,0]=ui
  for i in range(n):
    for j in range(1,i+1):
      g[i,j]=(g[j-1,j-1]-g[i,j-1])/((zi[i]-zi[j-1])*g[i,j-1])
  return np.diag(g)


def a(zi,ui):
  cdef int n
  n=len(zi)
  alist=np.zeros((n-1),dtype=np.double)
  for i in range(1,n):
    alist[i-1]=g(i,i,zi,ui)
  return alist

def g(int p,int i,zi,ui):
  if p==1:
    return ui[i-1]
  elif p>1:
    return (g(p-1,p-1,zi,ui)-g(p-1,i,zi,ui))/((zi[i-1]-zi[p-2])*g(p-1,i,zi,ui))

def A(zr,int n,zi,ui,a):
  if n==0:
    return 0.
  elif n==1:
    return a[0]
  else:
    return A(zr,n-1,zi,ui,a) + (zr-zi[n-2])*a[n-1]*A(zr,n-2,zi,ui,a)

def B(zr,int n,zi,ui,a):
  if n==0:
    return 1.
  elif n==1:
    return 1.
  else:
    return B(zr,n-1,zi,ui,a) + (zr-zi[n-2])*a[n-1]*B(zr,n-2,zi,ui,a)

def C(zr,zi,ui,a):
  cdef int n
  n=len(zi)
  return A(zr,n,zi,ui,a)/B(zr,n,zi,ui,a)





import numpy as np
import scipy.interpolate as interp
import scipy.integrate as integ


class Kernel(object):
    """This class handles the kernel of the analytic continuation."""

    def __init__(self, kind=None, re_axis=None, im_axis=None):
        """
        The kernel matrix is calculated from the real-frequency vector `re_axis`
        and the imaginary frequency or time vector `im_axis`.
        It is important to note that the Kernel does not know the inverse temperature
        beta. In case of the frequency kernels ('freq_fermionic' or 'freq_bosonic')
        the temperature is implicitly contained through the Matsubara frequencies.
        In case of the time kernels ('time_fermionic' or 'time_bosonic') the `im_axis`
        list has to be rescaled such that the temperature would be 1. Usually this is done
        by `imag_time = imag_time / beta`.

        For regular users of the library this is no problem, since the Kernel class is only
        instantiated inside the AnalyticContinuationProblem class, where the rescaling is done
        automatically. (Also the results are scaled back automatically there.)

        Parameters
        ----------
        kind : str, default=None
                  Which kind of kernel to use. Possible options:
                  'freq_bosonic', 'time_bosonic', 'freq_bosonic_xyz',
                  'freq_fermionic', 'time_fermionic', 'freq_fermionic_phsym',
                  'time_fermionic_phsym'
        re_axis : numpy.ndarray, default=None
                  real-frequency axis to generate the kernel matrix
        im_axis : numpy.ndarray, default=None
                  imaginary axis (time, Matsubara frequency)
                  to generate the kernel matrix
        """
        if (kind is None
                or re_axis is None
                or im_axis is None):
            raise ValueError('Kernel not correctly initialized')
        self.kind = kind
        self.re_axis = re_axis
        self.im_axis = im_axis
        self.original_matrix = self.kernel_matrix()
        self.matrix = np.copy(self.original_matrix)
        self.nw = self.re_axis.shape[0]
        self.niw = self.im_axis.shape[0]

    def kernel_matrix(self):
        """Compute the kernel matrix.
        If you want to implement another kernel,
        you just have to add another 'elif' here. """

        if self.kind == 'freq_bosonic':
            kernel = (self.re_axis ** 2)[None, :] \
                     / ((self.re_axis ** 2)[None, :]
                        + (self.im_axis ** 2)[:, None])
            WhereIsiwn0 = np.where(self.im_axis==0.0)[0]
            WhereIsw0 = np.where(self.re_axis==0.0)[0]
            if len(WhereIsiwn0==1) and len(WhereIsw0==1):
                kernel[WhereIsiwn0, WhereIsw0] = 1.0 # analytically with de l'Hospital
        elif self.kind == 'time_bosonic':
            kernel = 0.5 * self.re_axis[None, :] * (
                    np.exp(-self.re_axis[None, :] * self.im_axis[:, None])
                    + np.exp(-self.re_axis[None, :] * (1. - self.im_axis[:, None]))) / (
                                  1. - np.exp(-self.re_axis[None, :]))
            kernel[:, 0] = 1.  # analytically with de l'Hospital
        elif self.kind == 'freq_bosonic_xyz':
            kernel = -self.im_axis[:, None] / ((self.re_axis**2)[None, :] + (self.im_axis**2)[:, None])
            if self.im_axis[0] == 0.:
                kernel[0] = 0.
        elif self.kind == 'freq_fermionic':
            kernel = 1. / (1j * self.im_axis[:, None] - self.re_axis[None, :])
        elif self.kind == 'time_fermionic':
            kernel = np.exp(-self.im_axis[:, None] * self.re_axis[None, :]) \
                          / (1. + np.exp(-self.re_axis[None, :]))
        elif self.kind == 'freq_fermionic_phsym':  # in this case, the data must be purely real (the imaginary part!)
            print('Warning: phsym kernels do not give good results in this implementation. ')
            kernel = -2. * self.im_axis[:, None]\
                     / ((self.im_axis ** 2)[:, None]
                        + (self.re_axis ** 2)[None, :])
        elif self.kind == 'time_fermionic_phsym':
            print('Warning: phsym kernels do not give good results in this implementation. ')
            kernel = (np.cosh(self.im_axis[:, None] * self.re_axis[None, :])
                      + np.cosh((1. - self.im_axis[:, None]) * self.re_axis[None, :])) \
                     / (1. + np.cosh(self.re_axis[None, :]))
        else:
            raise ValueError("Unknown kernel type")
        return kernel

    def preblur(self, blur_width):
        """Preblur for the kernel, if applicable."""
        self.blur_width = blur_width
        if self.blur_width > 0. and (self.kind == 'freq_fermionic' or self.kind == 'freq_bosonic'):
            self.matrix = self.convolve_kernel()
        else:
            self.matrix = np.copy(self.original_matrix)

    def convolve_kernel(self):
        """
        Convolve bosonic or fermionic kernel with a Gaussian.

        The Gaussian is :math:`g(x) = \\frac{1}{b \\sqrt{2 \\pi}} exp(-x^2/(2b^2))`.
        In the fermionic case, the convolution can be written as
        :math:`K_{preblur}(i\\nu_n, \\omega) = \\int_{-5b}^{5b} dx\\; \\frac{g(x)}{i\\nu_n - x - \\omega}`

        In the bosonic case, the convolution can be written as
        :math:`K_{preblur}(i\\omega_n, \\nu) = \\frac{1}{2} \\int_{-5b}^{5b} dx\\; g(x) [\\frac{(x+\\nu)^2 }{ (x+\\nu)^2 + \\omega_n^2} + \\frac{(x-\\nu)^2 }{ (x-\\nu)^2 + \\omega_n^2}]`

        Integration over the Gaussian from :math:`-5b` to :math:`5b` is certainly sufficient.
        Thus the Gaussian has to be computed only once and integration by scipy.integrate.simps
        gives very accurate results even for tiny values of b.

        Note: More direct calculation by, e.g., scipy.integrate.quad is possible but unstable
        and extremely slow.
        """

        self.w_integration = np.linspace(-5. * self.blur_width, 5. * self.blur_width, num=201, endpoint=True)
        norm = 1. / (self.blur_width * np.sqrt(2. * np.pi))
        self.gaussian_numeric = norm * np.exp(-0.5 * (self.w_integration / self.blur_width) ** 2)

        if self.kind == 'freq_fermionic':
            integrand = self.gaussian_numeric[None, None, :] \
                        / (1j * self.im_axis[:, None, None] - self.re_axis[None, :, None] - self.w_integration[None, None, :])
        elif self.kind == 'freq_bosonic':
            WhereIsiwn0 = np.where(self.im_axis==0.0)[0]
            integrand_1 = self.gaussian_numeric[None, None, :] * (self.w_integration[None, None, :] + self.re_axis[None, :, None]) ** 2 \
                          / ((self.w_integration[None, None, :] + self.re_axis[None, :, None]) ** 2 + (self.im_axis[:, None, None]) ** 2)
            integrand_2 = self.gaussian_numeric[None, None, :] * (self.w_integration[None, None, :] - self.re_axis[None, :, None]) ** 2 \
                          / ((self.w_integration[None, None, :] - self.re_axis[None, :, None]) ** 2 + (self.im_axis[:, None, None]) ** 2)
            if len(WhereIsiwn0) > 0:  # analytically with de l'Hospital
                integrand_1[WhereIsiwn0] = self.gaussian_numeric[None, None, :]
                integrand_2[WhereIsiwn0] = self.gaussian_numeric[None, None, :]
            integrand = 0.5 * (integrand_1 + integrand_2)
        else:
            raise NotImplementedError('No preblur implemented for this kernel.')
        return integ.simps(integrand, x=self.w_integration, axis=-1)

    def blur(self, hidden_spectrum):
        """Convert hidden spectral function to spectral function."""
        if self.kind == 'freq_bosonic':
            h_interp = interp.InterpolatedUnivariateSpline(
                np.concatenate((-self.re_axis[:0:-1], self.re_axis)),
                np.concatenate((hidden_spectrum[:0:-1], hidden_spectrum)),
                ext='zeros')
        else:
            h_interp = interp.InterpolatedUnivariateSpline(
                self.re_axis,
                hidden_spectrum,
                ext='zeros')

        integrand = self.gaussian_numeric[None, :] * h_interp(self.re_axis[:, None] + self.w_integration[None, :])
        return integ.simps(integrand, x=self.w_integration, axis=-1)

    def real_matrix(self):
        """Return real and imaginary part one after another."""
        if self.kind == 'freq_fermionic':
            return np.concatenate((self.matrix.real, self.matrix.imag))
        else:
            return self.matrix

    def rotate_to_cov_eb(self, ucov):
        """Rotate first axis of kernel to eigenbasis of covariance matrix."""
        self.ucov = ucov
        self.matrix = np.dot(self.ucov.T.conj(), self.matrix)



import numpy as np
import h5py
import matplotlib.pyplot as plt
import scipy.interpolate as interp

import ana_cont.continuation as cont


class RealFrequencyGrid(object):
    """Class for real-frequency grids.

    Our real-frequency grids are always symmetric around 0.
    Thus they cover an interval [-wmax, wmax], where the endpoint is included.
    If the number of grid-points nw is an odd number, zero is included.
    We can create two types of grids: equispaced or non-equispaced (centered).
    The  latter is more dense in the low-frequency region, and more
    sparse in the high-frequency region.
    """

    def __init__(self, wmax=None, nw=None, type=None):
        """Initialize the real-frequency grid.

        wmax -- border of the real-frequency interval [-wmax, wmax]
        nw -- total number of real frequency grid points
        type -- grid type: 'equispaced grid' or 'centered grid'
        """
        self.wmax = wmax
        self.nw = nw
        self.type = type

    def update_wmax(self, wmax):
        print('update wmax to {}'.format(wmax))
        self.wmax = wmax

    def update_nw(self, nw):
        self.nw = nw

    def update_type(self, type):
        self.type = type

    def __str__(self):
        return 'real-frequency grid (wmax: {}, nw: {}, type: {})'.format(self.wmax, self.nw, self.type)

    def create_grid(self):
        """Create the real-frequency grid.

        There are four possible types of real-frequency grids.
        In each case, the endpoint is included in the grid.

        * An 'equispaced symmetric' grid is simply a linspace, containing also the endpoint.
        * The 'centered symmetric' grid is created by a tangent function,
          such that the grid points are denser around zero.
        * 'equispaced positive' is a simple linspace, starting from zero
          and containing only positive values.
        * 'centered positive' also starts from zero and contains only positive
          values, but close to zero they are lying denser.
        """
        if self.type == 'equispaced symmetric':
            self.grid = np.linspace(-self.wmax, self.wmax, num=self.nw, endpoint=True)
        elif self.type == 'centered symmetric':
            self.grid = self.wmax * np.tan(
                np.linspace(-np.pi / 2.1, np.pi / 2.1, num=self.nw, endpoint=True)) / np.tan(np.pi / 2.1)
        elif self.type == 'equispaced positive':
            self.grid = np.linspace(0., self.wmax, num=self.nw, endpoint=True)
        elif self.type == 'centered positive':
            self.grid = self.wmax * np.tan(
                np.linspace(0., np.pi / 2.1, num=self.nw, endpoint=True)) / np.tan(np.pi / 2.1)
        else:
            raise ValueError('Unknown real-frequency grid type.')
        print(self.grid)


def input_data_plot(mats, value, error, datatype, mom1=None, hartree=None):
    """Generate a very simple plot of the Matsubara data."""
    fig, ax = plt.subplots(ncols=1, nrows=1)
    ax.errorbar(mats, value.real,
                yerr=error, label='real part')
    ax.errorbar(mats, value.imag,
                yerr=error, label='imaginary part')
    if mom1 is not None:
        asymptote = mom1/mats
        astartind = np.searchsorted(asymptote,
                                    np.amin(value.imag))
        ax.plot(mats[astartind:], asymptote[astartind:], ':',
                label='asymptotic imaginary part', zorder=np.inf)
    ax.set_title('Input data')
    ax.set_xlabel('Matsubara frequency')
    ax.secondary_xaxis('top',
                       functions=(
                           interp.interp1d(mats,
                                           np.arange(mats.size),
                                           fill_value="extrapolate"),
                           interp.interp1d(np.arange(mats.size),
                                           mats,
                                           fill_value="extrapolate")
                       )).set_xlabel('Index')
    ax.set_ylabel('{}'.format(datatype))
    ymin, ymax = ax.get_ylim()
    xmin, xmax = ax.get_xlim()

    if datatype == "Self-energy":
        plt.text(xmin + 0.1 * (xmax - xmin), ymin + 0.9 * (ymax - ymin),
                 'Hartree energy = {:5.4f}'.format(hartree))
    plt.legend()
    plt.grid()
    plt.show()


class InputData(object):
    """Input data for the analytic continuation of a w2dynamics result."""

    def __init__(self, fname=None, iter_type=None, iter_num=None, data_type=None,
                 atom=None, orbital=None, spin=None, num_mats=None,
                 ignore_real_part=False):
        """Initialize the input data object.

        fname -- Name (str) of a valid w2dynamics DMFT output file.
            A w2dynamics version from 2020 or newer should be used.
        iter_type -- iteration type: 'dmft' or 'stat'
        iter_num -- iteration number: integer or 'last' (-1 also points to last)
        data_type -- "Green's function" or "Self-energy"
        atom -- which inequivalent atom (one-based integer), e.g. 1 leads to 'ineq-001'
        orbital -- which orbital component to load (one-based integer)
        spin -- which spin projection to load: 'up', 'down', 'average'
        num_mats -- number of Matsubara frequencies for continuation (integer)
        ignore_real_part -- if True, the real part is set to zero.
        """
        self.fname = fname
        self.iter_type = iter_type
        self.iter_num = iter_num
        self.data_type = data_type
        try:
            self.atom = int(atom)
        except ValueError:
            print('could not set atom')

        try:
            self.orbital = int(orbital)
        except ValueError:
            print('could not set orbital')

        self.spin = spin

        try:
            self.num_mats = int(num_mats)
        except ValueError:
            print('could not set num mats, using all available')
            self.num_mats = None

        if self.iter_type is None:
            self.iter_type = 'dmft'
        self.update_iter_num(self.iter_num)
        self.ignore_real_part = ignore_real_part

    def update_fname(self, fname):
        self.fname = fname

    def update_iter_type(self, iter_type):
        self.iter_type = iter_type.lower()
        self.get_iteration()

    def update_iter_num(self, iter_num):
        if iter_num is None or iter_num in ('', 'last', -1):
            self.iter_num = 'last'
        elif type(iter_num) == str or isinstance(iter_num, int):
            self.iter_num = '{:03}'.format(int(iter_num))
        else:
            raise ValueError('cannot read iteration number')
        self.get_iteration()

    def get_iteration(self):
        """Compose the whole iteration string, e.g. 'dmft-003'."""
        self.iteration = '{}-{}'.format(self.iter_type.lower(), self.iter_num)
        print('Iteration: {}'.format(self.iteration))

    def update_data_type(self, data_type):
        self.data_type = data_type
        if self.data_type == "Self-energy":
            pass
        elif self.data_type == "Green's function":
            pass
        else:
            raise ValueError('Unknown data type to read. Must be \'Self-energy\' or \'Green\'s function\'')

    def update_atom(self, atom):
        self.atom = int(atom)

    def update_orbital(self, orbital):
        self.orbital = int(orbital)

    def update_spin(self, spin):
        self.spin = spin

    def update_num_mats(self, num_mats):
        self.num_mats = int(num_mats)

    def load_data(self):
        """Load Matsubara-frequency data.

        Load w2dynamics data on Matsubara frequency axis.
        This is either a self-energy, or a Green's function.
        For the self-energy, there are two different possible formats.
        """
        self.generate_mats_freq()

        if self.data_type == "Self-energy":
            print('Try to load self-energy')
            f = h5py.File(self.fname, 'r')
            path_to_atom = '{}/ineq-{:03}/'.format(self.iteration, self.atom)
            all_keys = list(f[path_to_atom].keys())
            print(all_keys)
            f.close()
            if 'siw-full-jkerr' in all_keys:
                print('load siw 1')
                self.load_siw_1()
            elif 'siw-full' in all_keys:
                print('load siw 2')
                self.load_siw_2()
        elif self.data_type == "Green's function":
            self.load_giw()
        else:
            raise ValueError(
                'Unknown data type (Must be either \'Self-energy\' or \'Green\'s function\')')

        if self.ignore_real_part:
            self.value = 1j * self.value.imag

    def plot(self):
        input_data_plot(self.mats, self.value, self.error,
                        self.data_type, self.smom, self.hartree)

    def generate_mats_freq(self):
        """Generate the Matsubara frequency grid."""
        f = h5py.File(self.fname, 'r')
        beta = f['.config'].attrs['general.beta']
        if self.num_mats is None:
            self.num_mats = f['.config'].attrs['qmc.niw']
        f.close()
        self.mats = np.pi / beta * (2. * np.arange(self.num_mats) + 1.)

    def load_siw_1(self):
        """Load self-energy with jackknife error, type 1.

        In the first implementation, the self-energy with jackknife error
        was written in the group 'siw-full-jkerr', where the frequency
        axis is the first, followed by orbital and spin: (freq, orb, spin, orb, spin)
        The Hartree term is subtracted automatically.
        """
        path_to_atom = '{}/ineq-{:03}/'.format(self.iteration, self.atom)
        path_to_group = '{}/siw-full-jkerr/'.format(path_to_atom)
        path_to_value = path_to_group + 'value'
        path_to_error = path_to_group + 'error'
        path_to_smom = path_to_atom + 'smom-full/value'
        f = h5py.File(self.fname, 'r')
        if self.spin == 'up':
            data = f[path_to_value][:, self.orbital - 1, 0, self.orbital - 1, 0]
            err = f[path_to_error][:, self.orbital - 1, 0, self.orbital - 1, 0]
            smom = f[path_to_smom][self.orbital - 1, 0, self.orbital - 1, 0, 0]
            self.smom = f[path_to_smom][self.orbital - 1, 0, self.orbital - 1, 0, 1]
        elif self.spin == 'down':
            data = f[path_to_value][:, self.orbital - 1, 1, self.orbital - 1, 1]
            err = f[path_to_error][:, self.orbital - 1, 1, self.orbital - 1, 1]
            smom = f[path_to_smom][self.orbital - 1, 1, self.orbital - 1, 1, 0]
            self.smom = f[path_to_smom][self.orbital - 1, 1, self.orbital - 1, 1, 1]
        elif self.spin == 'average':
            data = 0.5 * (f[path_to_value][:, self.orbital - 1, 0, self.orbital - 1, 0]
                          + f[path_to_value][:, self.orbital - 1, 1, self.orbital - 1, 1])
            err = 1. / np.sqrt(2.) * (f[path_to_error][:, self.orbital - 1, 0, self.orbital - 1, 0]
                                     + f[path_to_error][:, self.orbital - 1, 1, self.orbital - 1, 1])
            smom = 0.5 * (f[path_to_smom][self.orbital - 1, 0, self.orbital - 1, 0, 0]
                          + f[path_to_smom][self.orbital - 1, 1, self.orbital - 1, 1, 0])
            self.smom = 0.5 * (f[path_to_smom][self.orbital - 1, 0, self.orbital - 1, 0, 1]
                               + f[path_to_smom][self.orbital - 1, 1, self.orbital - 1, 1, 1])
        f.close()
        niw = data.shape[0] // 2
        self.value = data[niw:niw + self.num_mats] - smom
        self.hartree = smom
        self.error = err[niw:niw + self.num_mats]

    def load_siw_2(self):
        """Load self-energy with jackknife error, type 2.

        In newer versions, the self-energy with jackknife error is stored
        in siw-full, in standard format (orb, spin, orb, spin, freq).
        If the QMC error is not present, it is set to a constant value err_const.
        The Hartree term is subtracted automatically.
        """
        path_to_atom = '{}/ineq-{:03}/'.format(self.iteration, self.atom)
        path_to_group = '{}/siw-full/'.format(path_to_atom)
        path_to_value = path_to_group + 'value'
        path_to_error = path_to_group + 'error'
        path_to_smom = path_to_atom + 'smom-full/value'
        err_const = 0.001
        f = h5py.File(self.fname, 'r')
        if self.spin == 'up':
            data = f[path_to_value][self.orbital - 1, 0, self.orbital - 1, 0]
            smom = f[path_to_smom][self.orbital - 1, 0, self.orbital - 1, 0, 0]
            self.smom = f[path_to_smom][self.orbital - 1, 0, self.orbital - 1, 0, 1]
            try:
                err = f[path_to_error][self.orbital - 1, 0, self.orbital - 1, 0]
            except KeyError:
                print('Warning: No self-energy error found; setting to constant {}'.format(err_const))
                err = np.ones_like(data) * err_const
        elif self.spin == 'down':
            data = f[path_to_value][self.orbital - 1, 1, self.orbital - 1, 1]
            smom = f[path_to_smom][self.orbital - 1, 1, self.orbital - 1, 1, 0]
            self.smom = f[path_to_smom][self.orbital - 1, 1, self.orbital - 1, 1, 1]
            try:
                err = f[path_to_error][self.orbital - 1, 1, self.orbital - 1, 1]
            except KeyError:
                print('Warning: No self-energy error found; setting to constant {}'.format(err_const))
                err = np.ones_like(data) * err_const
        elif self.spin == 'average':
            data = 0.5 * (f[path_to_value][self.orbital - 1, 0, self.orbital - 1, 0]
                          + f[path_to_value][self.orbital - 1, 1, self.orbital - 1, 1])
            smom = 0.5 * (f[path_to_smom][self.orbital - 1, 0, self.orbital - 1, 0, 0]
                          + f[path_to_smom][self.orbital - 1, 1, self.orbital - 1, 1, 0])
            self.smom = 0.5 * (f[path_to_smom][self.orbital - 1, 0, self.orbital - 1, 0, 1]
                               + f[path_to_smom][self.orbital - 1, 1, self.orbital - 1, 1, 1])
            try:
                err = 1. / np.sqrt(2.) * (f[path_to_error][self.orbital - 1, 0, self.orbital - 1, 0]
                                          + f[path_to_error][self.orbital - 1, 1, self.orbital - 1, 1])
            except KeyError:
                print('Warning: No self-energy error found; setting to constant {}'.format(err_const))
                err = np.ones_like(data) * err_const
        f.close()
        niw = data.shape[0] // 2
        self.value = data[niw:niw + self.num_mats] - smom
        self.hartree = smom
        self.error = err[niw:niw + self.num_mats]

    def load_giw(self):
        """Load Matsubara Green's function data.

        Here we have only one possible format, i.e. (orb, spin, orb, spin, freq).
        If no QMC error is present, we set it to a constant value err_const.
        """
        path_to_atom = '{}/ineq-{:03}/'.format(self.iteration, self.atom)
        path_to_group = '{}/giw-full/'.format(path_to_atom)
        path_to_value = path_to_group + 'value'
        path_to_error = path_to_group + 'error'

        err_const = 0.001
        f = h5py.File(self.fname, 'r')
        if self.spin == 'up':
            data = f[path_to_value][self.orbital - 1, 0, self.orbital - 1, 0]
            try:
                err = f[path_to_error][self.orbital - 1, 0, self.orbital - 1, 0]
            except KeyError:
                print('Warning: No Green\'s function error found; setting to constant {}'.format(err_const))
                err = np.ones_like(data) * err_const
        elif self.spin == 'down':
            data = f[path_to_value][self.orbital - 1, 1, self.orbital - 1, 1]
            try:
                err = f[path_to_error][self.orbital - 1, 1, self.orbital - 1, 1]
            except KeyError:
                print('Warning: No Green\'s function error found; setting to constant {}'.format(err_const))
                err = np.ones_like(data) * err_const
        elif self.spin == 'average':
            data = 0.5 * (f[path_to_value][self.orbital - 1, 0, self.orbital - 1, 0]
                          + f[path_to_value][self.orbital - 1, 1, self.orbital - 1, 1])
            try:
                err = 1. / np.sqrt(2.) * (f[path_to_error][self.orbital - 1, 0, self.orbital - 1, 0]
                                          + f[path_to_error][self.orbital - 1, 1, self.orbital - 1, 1])
            except KeyError:
                print('Warning: No Green\'s function error found; setting to constant {}'.format(err_const))
                err = np.ones_like(data) * err_const
        f.close()
        niw = data.shape[0] // 2
        self.value = data[niw:niw + self.num_mats]
        self.hartree = None
        self.smom = -1.
        self.error = err[niw:niw + self.num_mats]

class TextInputData(object):
    """Input data for analytic continuation, from a generic text file."""

    def __init__(self, fname=None, data_type=None,
                 num_mats=None, n_skip=None):
        """Initialize the text file input.

        fname -- file name of the text file
        data_type -- either "Green's function" or "Self-energy"
        num_mats -- number of Matsubara frequencies
        n_skip -- number of lines to skip at the beginning of the text file.

        Note the following points:
        * The Hartree term has to be subtracted beforehand,
            i.e. both the real and imaginary part of the data in the input file
            have to approach zero in the high-frequency limit,
        * The input file must contain only data on the positive half of the Matsubara axis,
        * data_type will not have any impact on the analytic continuation itself,
            but if it is a self-energy, the real part is calculated by Kramers-Kronig
            after the continuation, and the full self-energy is stored.
        * The Hartree term has to be handeled completely manually in pre- and postprocessing.
        """
        self.fname = fname
        self.data_type = data_type
        try:
            self.num_mats = int(num_mats)
        except ValueError:
            print('inferring number of Matsubaras from data')
            self.num_mats = None
        try:
            self.n_skip = int(n_skip)
        except ValueError:
            print('will not skip any lines')
            self.n_skip = 0

        self.atom = ''
        self.orbital = ''
        self.spin = ''

    def update_fname(self, fname):
        self.fname = fname

    def update_data_type(self, data_type):
        self.data_type = data_type

    def update_n_skip(self, n_skip):
        self.n_skip = n_skip

    def read_data(self):
        """Read the text file by np.loadtxt."""
        if self.data_type == 'bosonic':
            mats, val_re, err = np.loadtxt(self.fname, skiprows=self.n_skip, unpack=True)
            val_im = np.zeros_like(val_re)
        else:
            mats, val_re, val_im, err = np.loadtxt(self.fname, skiprows=self.n_skip, unpack=True)
        n_mats_data = mats.shape[0]
        if self.num_mats is None:
            self.num_mats = n_mats_data
        self.mats = mats[:self.num_mats]
        self.value = (val_re + 1j * val_im)[:self.num_mats]
        self.error = err[:self.num_mats]
        self.hartree = 0.

    def plot(self):
        input_data_plot(self.mats, self.value, self.error, self.data_type,
                        (-1 if self.data_type == "Green's function" else None),
                        self.hartree)


class OutputData(object):
    """Output data of the analytic continuation."""

    def __init__(self):
        pass

    def update_fname(self, fname):
        self.fname = fname
        print('Update output file name to: {}'.format(self.fname))

    def update(self, w, spec, input_data):
        """Update the output data with the latest input and results."""
        self.w = w
        self.spec = spec
        self.input_data = input_data

    def save(self, interpolate=False, n_reg=None):
        """Save output data to text file.

        Here we have to distinguish several different cases.
        * For a self-energy, we first do a Kramers-Kronig transformation
            to get also the real part.
            Then the Hartree term is added. Remember: when reading input data from a text file,
            the Hartree term has to be handeled manually, the program treats it as zero.
            The whole function, i.e. real and imaginary part of the self-energy, is saved.
        * In case of a Green's function, we save the spectral function,
            i.e. -Im[G(omega)] / pi. The real part is not computed.
        * If desired, an interpolation to a regular-spaced grid is done.

        Output format: text file.
        For spectral function: two colums: frequency, spectrum
        For self-energy: three columns: frequency, real part, imaginary part.
        """
        if self.input_data.data_type == "Self-energy":
            self.self_energy = cont.GreensFunction(spectrum=self.spec, wgrid=self.w, kind='fermionic').kkt()
            if interpolate:
                sw_real = self.interpolate(self.w, self.self_energy.real, n_reg)
                sw_imag = self.interpolate(self.w, self.self_energy.imag, n_reg)
                np.savetxt(self.fname, np.vstack((self.w_reg,
                                                  sw_real + self.input_data.hartree,
                                                  sw_imag)).T)
            else:
                np.savetxt(self.fname, np.vstack((self.w,
                                                  self.self_energy.real + self.input_data.hartree,
                                                  self.self_energy.imag)).T)

        elif (self.input_data.data_type == "Green's function"
              or self.input_data.data_type == "bosonic"):
            if interpolate:
                spec_interp = self.interpolate(self.w, self.spec, n_reg)
                np.savetxt(self.fname, np.vstack((self.w_reg, spec_interp)).T)
            else:
                np.savetxt(self.fname, np.vstack((self.w, self.spec)).T)
        else:
            print("Unknown data type {}, don't know how to save.".format(self.input_data.data_type))


    def interpolate(self, original_grid, original_function, n_reg):
        """Spline interpolation of real-frequency data."""
        self.w_reg = np.linspace(np.amin(self.w), np.amax(self.w), num=n_reg, endpoint=True)
        interp_function = interp.InterpolatedUnivariateSpline(original_grid, original_function)(self.w_reg)
        return interp_function



import sys
import numpy as np
if sys.version_info[0] > 2:
    from . import solvers
else:
    import solvers


class AnalyticContinuationProblem(object):
    """Specification of an analytic continuation problem.

    This This class is designed to hold the information that specifies the analytic continuation problem:

    * Imaginary axis of the data
    * Real axis on which the result is anticipated
    * Imaginary-axis data
    * Type of kernel (fermionic, bosonic, time, frequency)
    * Inverse temperature beta (only necessary for time kernels)

    Furthermore the class provides an interface to call the solver.
    """
    def __init__(self, im_axis=None, re_axis=None,
                 im_data=None, kernel_mode=None, beta=None):
        self.kernel_mode = kernel_mode
        self.im_axis = im_axis
        self.re_axis = re_axis
        self.im_data = im_data
        if self.kernel_mode == 'freq_bosonic':
            pass # not necessary to do anything additionally here
        elif self.kernel_mode == 'time_bosonic':
            self.im_axis = im_axis/beta
            self.re_axis = re_axis*beta
            self.im_data = im_data*beta
            self.beta = beta
        elif self.kernel_mode == 'freq_fermionic':
            pass
            #self.im_data = np.concatenate((im_data.real, im_data.imag))
        elif self.kernel_mode == 'freq_fermionic_phsym':
            # check if data are purely imaginary
            if np.allclose(im_axis.real, 0.):
                self.im_data = im_data.imag
            elif np.allclose(im_axis.imag, 0.): # if only the imaginary part is passed
                self.im_data = im_data.real
            else:
                print('The data are neither purely real nor purely imaginary,')
                print('you cannot use a ph-symmetric kernel in this case.')
                sys.exit()
        elif self.kernel_mode == 'time_fermionic' \
                or self.kernel_mode == 'time_fermionic_phsym':
            self.im_axis = im_axis / beta
            self.re_axis = re_axis * beta
            self.im_data = im_data
            self.beta = beta
        else:
            raise ValueError("Unsupported kernel_mode.")


    def solve(self, method='', **kwargs):
        """ Interface function for solving the analytic continuation problem.

        :param method: Analytic continuation method, possible choices are 'maxent_svd', 'pade'
        :param kwargs: Further keyword arguments, specific to the solution method
        :return: OptimizationResult object

        This function first creates an instance of an AnalyticContinuationSolver object,
        then the respective solve function is called.
        """

        if method == 'maxent_svd':
            self.solver = solvers.MaxentSolverSVD(
                self.im_axis, self.re_axis, self.im_data,
                kernel_mode = self.kernel_mode, 
                **kwargs)
            sol = self.solver.solve(**kwargs)
            # TODO implement a postprocessing method, where the following should be done more carefully
            if self.kernel_mode == 'time_fermionic':
                sol[0].A_opt *= self.beta
            elif self.kernel_mode == 'time_bosonic':
                sol[0].A_opt *= self.beta
                sol[0].backtransform /= self.beta
            return sol
        elif method == 'maxent_mc':
            raise NotImplementedError
        elif method == 'pade':
            self.solver = solvers.PadeSolver(
                im_axis = self.im_axis, re_axis = self.re_axis,
                im_data = self.im_data)
            return self.solver.solve()
        else:
            raise ValueError("Unknown continuation method. Use one of 'maxent_svd', 'pade'.")

    def partial_solution(self, method='', **kwargs):
        """Maxent optimization at a specific value of alpha."""
        if method == 'maxent_svd':
            self.solver = solvers.MaxentSolverSVD(
                self.im_axis, self.re_axis, self.im_data,
                kernel_mode=self.kernel_mode,
                **kwargs)

            # truncate ustart to actual number of significant singular values
            kwargs['ustart'] = kwargs['ustart'][:self.solver.n_sv]
            sol = self.solver.maxent_optimization(**kwargs)
            if self.kernel_mode == 'time_fermionic':
                sol.A_opt *= self.beta
            elif self.kernel_mode == 'time_bosonic':
                sol.A_opt *= self.beta
                sol.backtransform /= self.beta
            return sol
        else:
            return ValueError("Unknown solver method.")


class GreensFunction(object):
    """
    This class defines a GreensFunction object. The main use of this
    is to calculate a full Green's function with real- and imaginary part
    from a spectrum.
    """

    def __init__(self, spectrum = None, wgrid = None, kind = ''):
        """Initialize with spectral function and real-frequency grid.
        """
        self.spectrum = spectrum
        self.wgrid = wgrid
        self.wmin = self.wgrid[0]
        self.wmax = self.wgrid[-1]
        self.dw = np.diff(
            np.concatenate(([self.wmin],
                            (self.wgrid[1:] + self.wgrid[:-1])/2.,
                            [self.wmax])))
        self.kind = kind # fermionic_phsym, bosonic,       fermionic
        #            or: symmetric,       antisymmetric, general

    def kkt(self):
        """Kramers Kronig transformation

        Obtain full complex Green's function from its imaginary part.
        """
        if self.kind == 'fermionic_phsym' or self.kind == 'symmetric':
            if self.wmin < 0.:
                print('warning: wmin<0 not permitted for fermionic_phsym greens functions.')

            m = 2. * self.dw[:,None] * self.wgrid[:,None] * self.spectrum[:,None] \
                / (self.wgrid[None,:]**2 - self.wgrid[:,None]**2)

        elif self.kind == 'bosonic' or self.kind == 'antisymmetric':
            if self.wmin < 0.:
                print('warning: wmin<0 not permitted for bosonic (antisymmetric) spectrum.')
            m = 2. * self.dw[:,None] * self.wgrid[None,:] * self.spectrum[:,None]\
                /(self.wgrid[None,:]**2 - self.wgrid[:,None]**2)
        
        elif self.kind == 'fermionic' or self.kind == 'general':
            m = self.dw[:,None] * self.spectrum[:,None]\
                /(self.wgrid[None,:]-self.wgrid[:,None])

        else:
            raise ValueError("Unknown kind of Greens function.")

        np.fill_diagonal(m, 0.)  # set manually where w==w'
        self.g_real = np.sum(m, axis=0)
        self.g_imag = -self.spectrum*np.pi
        return self.g_real + 1j*self.g_imag



import os
import numpy as np
import matplotlib.pyplot as plt
from PyQt5 import QtWidgets

import ana_cont.continuation as cont
from gui.gui_backend import RealFrequencyGrid, InputData, TextInputData, OutputData
from gui.maxent_ui import Ui_MainWindow


class MainWindow(QtWidgets.QMainWindow, Ui_MainWindow):
    """The Main Window of the graphical user interface.

    The class MainWindow inherits from Ui_MainWindow, which is
    defined in maxent_ui.py. The latter file is autogenerated
    by pyuic from maxent_ui.ui [`pyuic5 maxent_ui.ui -o maxent_ui.py`]
    The ui file can be edited by the QtDesigner.
    """

    def __init__(self, *args, obj=None, **kwargs):
        """Connect the widgets, instantiate the main classes."""
        super(MainWindow, self).__init__(*args, **kwargs)
        self.setupUi(self)

        # real-frequency grid
        self.realgrid = RealFrequencyGrid(wmax=float(self.max_real_freq.text()),
                                          nw=int(self.num_real_freq.text()),
                                          type=str(self.grid_type_combo.currentText()))
        self.connect_realgrid_button()
        self.connect_wmax()
        self.connect_nw()
        self.connect_grid_type()

        # input data
        self.input_data = InputData(fname=str(self.inp_file_name.text()),
                                    iter_type=str(self.iteration_type_combo.currentText()),
                                    iter_num=str(self.iteration_number.text()),
                                    data_type=str(self.inp_data_type.currentText()),
                                    atom=str(self.atom_number.text()),
                                    orbital=str(self.orbital_number.text()),
                                    spin=str(self.spin_type_combo.currentText()),
                                    num_mats=str(self.num_mats_freq.text()))
        self.connect_select_button()
        self.connect_load_button()
        self.connect_show_button()
        self.connect_load_button_text()
        self.connect_show_button_2()
        self.connect_select_button_2()

        # text display field and "Do it" button
        self.text_output.setReadOnly(True)
        self.connect_doit_button()

        # output data
        self.output_data = OutputData()
        self.connect_select_output_button()
        self.connect_save_button()


    def connect_realgrid_button(self):
        self.gen_real_grid_button.clicked.connect(lambda: self.realgrid.create_grid())

    def connect_wmax(self):
        self.max_real_freq.returnPressed.connect(
            lambda: self.realgrid.update_wmax(float(self.max_real_freq.text())))
        self.max_real_freq.editingFinished.connect(
            lambda: self.realgrid.update_wmax(float(self.max_real_freq.text())))

    def connect_nw(self):
        self.num_real_freq.returnPressed.connect(
            lambda: self.realgrid.update_nw(int(self.num_real_freq.text())))
        self.num_real_freq.editingFinished.connect(
            lambda: self.realgrid.update_nw(int(self.num_real_freq.text())))

    def connect_grid_type(self):
        self.grid_type_combo.activated.connect(
            lambda: self.realgrid.update_type(str(self.grid_type_combo.currentText()))
        )

    def preset_fnames(self, fname):
        self.inp_file_name.setText(fname)
        self.inp_file_name_2.setText(fname)

    def connect_fname_input(self):
        self.inp_file_name.editingFinished.connect(
            lambda: self.input_data.update_fname(str(self.inp_file_name.text())))
        self.inp_file_name.textChanged.connect(
            lambda: self.input_data.update_fname(str(self.inp_file_name.text())))

    def get_fname(self):
        self.inp_file_name.setText(
            QtWidgets.QFileDialog.getOpenFileName(self,
                    'Open file', os.getcwd(), "HDF5 files (*.hdf5)")[0])

    def connect_select_button(self):
        self.select_file_button.clicked.connect(self.get_fname)

    def get_fname_text(self):
        self.inp_file_name_2.setText(
            QtWidgets.QFileDialog.getOpenFileName(self,
                    'Open file', os.getcwd(), "text files (*.dat *.txt)")[0])

    def connect_select_button_2(self):
        self.select_file_button_2.clicked.connect(self.get_fname_text)

    def connect_data_type(self):
        self.inp_data_type.activated.connect(
            lambda: self.input_data.update_data_type(str(self.inp_data_type.currentText())))

    def connect_iteration_type(self):
        self.iteration_type_combo.activated.connect(
            lambda: self.input_data.update_iter_type(str(self.iteration_type_combo.currentText())))

    def connect_iteration_number(self):
        self.iteration_number.editingFinished.connect(
            lambda: self.input_data.update_iter_num(str(self.iteration_number.text())))

    def connect_atom(self):
        self.atom_number.editingFinished.connect(
            lambda: self.input_data.update_atom(int(self.atom_number.text())))

    def connect_orbital(self):
        self.orbital_number.editingFinished.connect(
            lambda: self.input_data.update_orbital(int(self.orbital_number.text())))

    def connect_spin(self):
        self.spin_type_combo.activated.connect(
            lambda: self.input_data.update_spin(str(self.spin_type_combo.currentText())))

    def connect_num_mats(self):
        self.num_mats_freq.editingFinished.connect(
            lambda: self.input_data.update_num_mats(int(self.num_mats_freq.text()))
        )

    def connect_show_button(self):
        self.show_data_button.clicked.connect(
            lambda: self.input_data.plot()
        )

    def connect_show_button_2(self):
        self.show_data_button_2.clicked.connect(
            lambda: self.input_data.plot()
        )

    def load_w2dynamics_data(self):
        self.input_data = InputData(fname=str(self.inp_file_name.text()),
                                    iter_type=str(self.iteration_type_combo.currentText()),
                                    iter_num=str(self.iteration_number.text()),
                                    data_type=str(self.inp_data_type.currentText()),
                                    atom=str(self.atom_number.text()),
                                    orbital=str(self.orbital_number.text()),
                                    spin=str(self.spin_type_combo.currentText()),
                                    num_mats=str(self.num_mats_freq.text()),
                                    ignore_real_part=self.ignore_checkbox.isChecked())
        self.input_data.load_data()

    def connect_load_button(self):
        self.load_data_button.clicked.connect(self.load_w2dynamics_data)

    def load_text_data(self):
        self.input_data = TextInputData(fname=str(self.inp_file_name_2.text()),
                                        data_type=str(self.inp_data_type_text.currentText()),
                                        n_skip=str(self.n_skip.text()),
                                        num_mats=str(self.num_mats_freq_text.text()))
        self.input_data.read_data()

    def connect_load_button_text(self):
        self.load_data_button_2.clicked.connect(self.load_text_data)

    def get_preblur(self):
        preblur_checked = self.preblur_checkbox.isChecked()
        try:
            bw = float(self.blur_width.text()) if preblur_checked else 0.
        except ValueError:
            print('Invalid input for blur width, setting to 0.')
            bw = 0.
        preblur = preblur_checked and bw > 0.
        return preblur, bw

    def main_function(self):
        """Main function for the analytic continuation procedure.

        This function is called when the "Do it" button is clicked.
        It performs an analytical continuation for the present settings
        and shows a plot.
        """
        self.ana_cont_probl = cont.AnalyticContinuationProblem(im_axis=self.input_data.mats,
                                                               im_data=self.input_data.value,
                                                               re_axis=self.realgrid.grid,
                                                               kernel_mode='freq_fermionic')
        model = np.ones_like(self.realgrid.grid)
        model /= np.trapz(model, self.realgrid.grid)

        preblur, bw = self.get_preblur()

        sol = self.ana_cont_probl.solve(method='maxent_svd',
                                        optimizer='newton',
                                        alpha_determination='chi2kink',
                                        model=model,
                                        stdev=self.input_data.error,
                                        interactive=False, alpha_start=1e14, alpha_end=1e-3,
                                        preblur=preblur, blur_width=bw)

        inp_str = 'atom {}, orb {}, spin {}, blur {}: '.format(self.input_data.atom,
                                                               self.input_data.orbital,
                                                               self.input_data.spin,
                                                               bw)
        all_chis = np.isfinite(np.array([s.chi2 for s in sol[1]]))
        res_str = 'alpha_opt={:3.2f}, chi2(alpha_opt)={:3.2f}, min(chi2)={:3.2f}'.format(
            sol[0].alpha, sol[0].chi2, np.amin(all_chis)
        )
        self.text_output.append(inp_str + res_str)
        alphas = [s.alpha for s in sol[1]]
        chis = [s.chi2 for s in sol[1]]

        self.output_data.update(self.realgrid.grid, sol[0].A_opt, self.input_data)

        fig, ax = plt.subplots(ncols=2, nrows=2, figsize=(11.75, 8.25))  # A4 paper size
        ax[0, 0].loglog(alphas, chis, marker='s', color='black')
        ax[0, 0].loglog(sol[0].alpha, sol[0].chi2, marker='*', color='red', markersize=15)
        ax[0, 0].set_xlabel(r'$\alpha$')
        ax[0, 0].set_ylabel(r'$\chi^2(\alpha)$')

        ax[1, 0].plot(self.realgrid.grid, sol[0].A_opt)
        ax[1, 0].set_xlabel(r'$\omega$')
        ax[1, 0].set_ylabel('spectrum')

        ax[0, 1].plot(self.input_data.mats, self.input_data.value.real,
                      color='blue', ls=':', marker='x', markersize=5,
                      label='Re[data]')
        ax[0, 1].plot(self.input_data.mats, self.input_data.value.imag,
                      color='green', ls=':', marker='+', markersize=5,
                      label='Im[data]')
        ax[0, 1].plot(self.input_data.mats, sol[0].backtransform.real,
                      ls='--', color='gray', label='Re[fit]')
        ax[0, 1].plot(self.input_data.mats, sol[0].backtransform.imag,
                      color='gray', label='Im[fit]')
        ax[0, 1].set_xlabel(r'$\nu_n$')
        ax[0, 1].set_ylabel(self.input_data.data_type)
        ax[0, 1].legend()

        ax[1, 1].plot(self.input_data.mats, (self.input_data.value - sol[0].backtransform).real,
                      ls='--', label='real part')
        ax[1, 1].plot(self.input_data.mats, (self.input_data.value - sol[0].backtransform).imag,
                      label='imaginary part')
        ax[1, 1].set_xlabel(r'$\nu_n$')
        ax[1, 1].set_ylabel('data $-$ fit')
        ax[1, 1].legend()
        plt.tight_layout()
        plt.show()

    def connect_doit_button(self):
        self.doit_button.clicked.connect(lambda: self.main_function())

    def connect_fname_output(self):
        self.out_file_name.editingFinished.connect(
            lambda: self.output_data.update_fname(str(self.out_file_name.text())))
        self.inp_file_name.textChanged.connect(
            lambda: self.output_data.update_fname(str(self.out_file_name.text())))

    def get_fname_output(self):
        fname_out = QtWidgets.QFileDialog.getSaveFileName(self,
                    'Save as', '/'.join(self.input_data.fname.split('/')[:-1]), "DAT files (*.dat)")[0]
        self.out_file_name.setText(fname_out)
        self.output_data.update_fname(fname_out)

    def connect_select_output_button(self):
        self.output_directory_button.clicked.connect(self.get_fname_output)

    def save_output(self):
        fname_out = str(self.out_file_name.text())
        if fname_out == '':
            print('Error in saving: First you have to specify the output file name.')
            return 1

        self.output_data.update_fname(fname_out)
        try:
            self.output_data.save(interpolate=self.interpolate_checkbox.isChecked(),
                                 n_reg=int(self.n_interpolation.text()))
        except AttributeError:
            print('Error in saving: First you have to specify the output file name.')

    def connect_save_button(self):
        self.save_button.clicked.connect(lambda: self.save_output())




# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'maxent_bosonic_ui.ui'
#
# Created by: PyQt5 UI code generator 5.9.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(759, 629)
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.real_freq_frame = QtWidgets.QFrame(self.centralwidget)
        self.real_freq_frame.setGeometry(QtCore.QRect(10, 10, 231, 171))
        self.real_freq_frame.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.real_freq_frame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.real_freq_frame.setObjectName("real_freq_frame")
        self.label = QtWidgets.QLabel(self.real_freq_frame)
        self.label.setGeometry(QtCore.QRect(10, 10, 141, 17))
        self.label.setObjectName("label")
        self.label_3 = QtWidgets.QLabel(self.real_freq_frame)
        self.label_3.setGeometry(QtCore.QRect(10, 70, 31, 17))
        self.label_3.setObjectName("label_3")
        self.label_4 = QtWidgets.QLabel(self.real_freq_frame)
        self.label_4.setGeometry(QtCore.QRect(10, 110, 21, 17))
        self.label_4.setObjectName("label_4")
        self.grid_type_combo = QtWidgets.QComboBox(self.real_freq_frame)
        self.grid_type_combo.setGeometry(QtCore.QRect(10, 40, 201, 25))
        self.grid_type_combo.setObjectName("grid_type_combo")
        self.grid_type_combo.addItem("")
        self.grid_type_combo.addItem("")
        self.max_real_freq = QtWidgets.QLineEdit(self.real_freq_frame)
        self.max_real_freq.setGeometry(QtCore.QRect(40, 70, 41, 25))
        self.max_real_freq.setObjectName("max_real_freq")
        self.num_real_freq = QtWidgets.QLineEdit(self.real_freq_frame)
        self.num_real_freq.setGeometry(QtCore.QRect(40, 110, 41, 25))
        self.num_real_freq.setObjectName("num_real_freq")
        self.gen_real_grid_button = QtWidgets.QPushButton(self.real_freq_frame)
        self.gen_real_grid_button.setGeometry(QtCore.QRect(90, 110, 71, 25))
        self.gen_real_grid_button.setObjectName("gen_real_grid_button")
        self.input_data_tabs = QtWidgets.QTabWidget(self.centralwidget)
        self.input_data_tabs.setGeometry(QtCore.QRect(250, 10, 501, 171))
        self.input_data_tabs.setObjectName("input_data_tabs")
        self.text_tab = QtWidgets.QWidget()
        self.text_tab.setObjectName("text_tab")
        self.label_14 = QtWidgets.QLabel(self.text_tab)
        self.label_14.setGeometry(QtCore.QRect(10, 10, 221, 17))
        self.label_14.setObjectName("label_14")
        self.label_16 = QtWidgets.QLabel(self.text_tab)
        self.label_16.setGeometry(QtCore.QRect(10, 40, 67, 17))
        self.label_16.setObjectName("label_16")
        self.inp_file_name_2 = QtWidgets.QLineEdit(self.text_tab)
        self.inp_file_name_2.setGeometry(QtCore.QRect(90, 40, 361, 25))
        self.inp_file_name_2.setObjectName("inp_file_name_2")
        self.select_file_button_2 = QtWidgets.QToolButton(self.text_tab)
        self.select_file_button_2.setGeometry(QtCore.QRect(460, 40, 26, 24))
        self.select_file_button_2.setObjectName("select_file_button_2")
        self.load_data_button_2 = QtWidgets.QPushButton(self.text_tab)
        self.load_data_button_2.setGeometry(QtCore.QRect(300, 110, 89, 25))
        self.load_data_button_2.setObjectName("load_data_button_2")
        self.show_data_button_2 = QtWidgets.QPushButton(self.text_tab)
        self.show_data_button_2.setGeometry(QtCore.QRect(400, 110, 89, 25))
        self.show_data_button_2.setObjectName("show_data_button_2")
        self.label_17 = QtWidgets.QLabel(self.text_tab)
        self.label_17.setGeometry(QtCore.QRect(10, 80, 31, 17))
        self.label_17.setObjectName("label_17")
        self.n_skip = QtWidgets.QLineEdit(self.text_tab)
        self.n_skip.setGeometry(QtCore.QRect(50, 80, 31, 25))
        self.n_skip.setObjectName("n_skip")
        self.label_18 = QtWidgets.QLabel(self.text_tab)
        self.label_18.setGeometry(QtCore.QRect(90, 80, 151, 17))
        self.label_18.setObjectName("label_18")
        self.label_19 = QtWidgets.QLabel(self.text_tab)
        self.label_19.setGeometry(QtCore.QRect(10, 110, 31, 17))
        self.label_19.setObjectName("label_19")
        self.num_mats_freq_text = QtWidgets.QLineEdit(self.text_tab)
        self.num_mats_freq_text.setGeometry(QtCore.QRect(50, 110, 31, 25))
        self.num_mats_freq_text.setObjectName("num_mats_freq_text")
        self.label_20 = QtWidgets.QLabel(self.text_tab)
        self.label_20.setGeometry(QtCore.QRect(90, 110, 161, 17))
        self.label_20.setObjectName("label_20")
        self.input_data_tabs.addTab(self.text_tab, "")
        self.continuation_frame = QtWidgets.QFrame(self.centralwidget)
        self.continuation_frame.setGeometry(QtCore.QRect(10, 190, 741, 391))
        self.continuation_frame.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.continuation_frame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.continuation_frame.setObjectName("continuation_frame")
        self.doit_button = QtWidgets.QPushButton(self.continuation_frame)
        self.doit_button.setGeometry(QtCore.QRect(590, 20, 131, 41))
        self.doit_button.setObjectName("doit_button")
        self.blur_width = QtWidgets.QLineEdit(self.continuation_frame)
        self.blur_width.setGeometry(QtCore.QRect(80, 40, 113, 25))
        self.blur_width.setObjectName("blur_width")
        self.label_11 = QtWidgets.QLabel(self.continuation_frame)
        self.label_11.setGeometry(QtCore.QRect(30, 40, 51, 17))
        self.label_11.setObjectName("label_11")
        self.text_output = QtWidgets.QTextEdit(self.continuation_frame)
        self.text_output.setGeometry(QtCore.QRect(30, 80, 691, 231))
        self.text_output.setObjectName("text_output")
        self.save_button = QtWidgets.QPushButton(self.continuation_frame)
        self.save_button.setGeometry(QtCore.QRect(630, 360, 89, 25))
        self.save_button.setObjectName("save_button")
        self.output_directory_button = QtWidgets.QToolButton(self.continuation_frame)
        self.output_directory_button.setGeometry(QtCore.QRect(120, 360, 26, 24))
        self.output_directory_button.setObjectName("output_directory_button")
        self.out_file_name = QtWidgets.QLineEdit(self.continuation_frame)
        self.out_file_name.setGeometry(QtCore.QRect(160, 360, 451, 25))
        self.out_file_name.setObjectName("out_file_name")
        self.label_12 = QtWidgets.QLabel(self.continuation_frame)
        self.label_12.setGeometry(QtCore.QRect(30, 360, 91, 17))
        self.label_12.setObjectName("label_12")
        self.n_interpolation = QtWidgets.QLineEdit(self.continuation_frame)
        self.n_interpolation.setGeometry(QtCore.QRect(200, 320, 41, 25))
        self.n_interpolation.setObjectName("n_interpolation")
        self.label_13 = QtWidgets.QLabel(self.continuation_frame)
        self.label_13.setGeometry(QtCore.QRect(250, 320, 201, 17))
        self.label_13.setObjectName("label_13")
        self.preblur_checkbox = QtWidgets.QCheckBox(self.continuation_frame)
        self.preblur_checkbox.setGeometry(QtCore.QRect(30, 10, 92, 23))
        self.preblur_checkbox.setObjectName("preblur_checkbox")
        self.interpolate_checkbox = QtWidgets.QCheckBox(self.continuation_frame)
        self.interpolate_checkbox.setGeometry(QtCore.QRect(30, 320, 171, 23))
        self.interpolate_checkbox.setObjectName("interpolate_checkbox")
        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(MainWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 759, 22))
        self.menubar.setObjectName("menubar")
        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(MainWindow)
        self.statusbar.setObjectName("statusbar")
        MainWindow.setStatusBar(self.statusbar)

        self.retranslateUi(MainWindow)
        self.input_data_tabs.setCurrentIndex(0)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "MainWindow"))
        self.real_freq_frame.setToolTip(_translate("MainWindow", "configure real-frequency grid"))
        self.real_freq_frame.setWhatsThis(_translate("MainWindow", "real-frequency grid"))
        self.label.setText(_translate("MainWindow", "Real-frequency grid"))
        self.label_3.setText(_translate("MainWindow", "max"))
        self.label_4.setText(_translate("MainWindow", "n"))
        self.grid_type_combo.setToolTip(_translate("MainWindow", "equispaced or centered grid (denser around Fermi energy)"))
        self.grid_type_combo.setItemText(0, _translate("MainWindow", "equispaced positive"))
        self.grid_type_combo.setItemText(1, _translate("MainWindow", "centered positive"))
        self.max_real_freq.setToolTip(_translate("MainWindow", "upper border of real-frequency grid. (lower border is set symmetrically)"))
        self.max_real_freq.setText(_translate("MainWindow", "20"))
        self.num_real_freq.setToolTip(_translate("MainWindow", "number frequencies on real axis; should be an odd number"))
        self.num_real_freq.setText(_translate("MainWindow", "401"))
        self.gen_real_grid_button.setText(_translate("MainWindow", "Generate"))
        self.label_14.setText(_translate("MainWindow", "Load susceptibility from text file"))
        self.label_16.setText(_translate("MainWindow", "file name"))
        self.inp_file_name_2.setToolTip(_translate("MainWindow", "file path and name of a w2dynamics output file"))
        self.select_file_button_2.setToolTip(_translate("MainWindow", "choose an input file"))
        self.select_file_button_2.setText(_translate("MainWindow", "..."))
        self.load_data_button_2.setText(_translate("MainWindow", "Load data"))
        self.show_data_button_2.setToolTip(_translate("MainWindow", "click this if you want to plot the data after loading"))
        self.show_data_button_2.setText(_translate("MainWindow", "Show data"))
        self.label_17.setText(_translate("MainWindow", "Skip"))
        self.label_18.setText(_translate("MainWindow", "lines at the beginning"))
        self.label_19.setText(_translate("MainWindow", "Use"))
        self.label_20.setText(_translate("MainWindow", "Matsubara frequencies"))
        self.input_data_tabs.setTabText(self.input_data_tabs.indexOf(self.text_tab), _translate("MainWindow", "text file"))
        self.doit_button.setToolTip(_translate("MainWindow", "perform the analytical continuation"))
        self.doit_button.setText(_translate("MainWindow", "Do it!"))
        self.blur_width.setToolTip(_translate("MainWindow", "set the blur width here"))
        self.blur_width.setText(_translate("MainWindow", "0.1"))
        self.label_11.setText(_translate("MainWindow", "Width"))
        self.text_output.setToolTip(_translate("MainWindow", "in this field some output will be shown"))
        self.save_button.setToolTip(_translate("MainWindow", "click this button to save the output"))
        self.save_button.setText(_translate("MainWindow", "Save"))
        self.output_directory_button.setToolTip(_translate("MainWindow", "Choose a directory, where you want to save the output"))
        self.output_directory_button.setText(_translate("MainWindow", "..."))
        self.out_file_name.setToolTip(_translate("MainWindow", "type full output name here (including path)"))
        self.label_12.setText(_translate("MainWindow", "Output file:"))
        self.n_interpolation.setToolTip(_translate("MainWindow", "number of regularly spaced grid points for interpolation"))
        self.n_interpolation.setText(_translate("MainWindow", "0"))
        self.label_13.setText(_translate("MainWindow", "regularly spaced grid points"))
        self.preblur_checkbox.setToolTip(_translate("MainWindow", "check this if you want to use preblur"))
        self.preblur_checkbox.setText(_translate("MainWindow", "Preblur"))
        self.interpolate_checkbox.setToolTip(_translate("MainWindow", "check this for interpolating output to regular grid"))
        self.interpolate_checkbox.setText(_translate("MainWindow", "Interpolate output to"))




import sys
import numpy as np
import scipy.optimize as opt
import collections
from abc import ABC, abstractmethod
from . import kernels

if sys.version_info[0] > 2:
    try:
        from . import pade
    except:
        pass
else:
    try:
        import pade
    except:
        pass


class AnalyticContinuationSolver(ABC):
    """Abstract base class for solver classes.

    Each Solver class has to inherit from this class.
    The purpose of this is to ensure that each child class
    has a method 'solve' implemented.
    """
    @abstractmethod
    def solve(self):
        pass


class PadeSolver(AnalyticContinuationSolver):
    """Pade solver"""
    def __init__(self, im_axis, re_axis, im_data):
        """
        Parameters
        ----------
        im_axis : numpy.ndarray
                  Matsubara frequencies which are used for the continuation
        re_axis : numpy.ndarray
                  Real-frequency points at which the Pade interpolant is evaluated
        im_data : Green's function values at the given points `im_axis`
        """
        self.im_axis = im_axis
        self.re_axis = re_axis
        self.im_data = im_data

        # Compute the Pade coefficients
        self.a = pade.compute_coefficients(1j * self.im_axis, self.im_data)

    def check(self, im_axis_fine=None):
        """Sanity check for Pade approximant

        Evaluate the Pade approximant on the imaginary axis,
        however not only at Matsubara frequencies, but on a
        dense grid. If the approximant is good, then this
        should yield a smooth interpolating curve on the Matsubara
        axis. On the other hand, little 'spikes' are a hint
        for pole-zero pairs close to the imaginary axis. This
        is usually a result of noise and a different choice of
        Matsubara frequencies may help.

        Parameters
        ----------
        im_axis_fine : numpy.ndarray, default=None
                  Imaginary-axis points where the approximant is
                  evaluated for checking. If not specified,
                  an array of length 500 is generated, which goes
                  up to twice the maximum frequency that was used
                  for constructing the approximant.

        Returns
        -------
        numpy.ndarray
                  Values of the Pade approximant at the points
                  of `im_axis_fine`.
        """

        if im_axis_fine is None:
            self.ivcheck = np.linspace(0, 2 * np.max(self.im_axis), num=500)
        else:
            self.ivcheck = im_axis_fine
        check = pade.C(self.ivcheck * 1j, 1j * self.im_axis,
                            self.im_data, self.a)
        return check

    def solve(self, show_plot=False):
        """Compute the Pade approximation on the real axis.

        The main numerically heavy computation is done in the Cython module
        `pade.pyx`. Here we just call the functions.
        In the Pade method, the numerator and denominator approximants
        are generated separately, and then the division is done.
        As an additional feature we add the callable `numerator_function`
        and `denominator_function` to the OptimizationResult object.
        """

        def numerator_function(z):
            return pade.A(z, self.im_axis.shape[0],
                           1j * self.im_axis, self.im_data, self.a)

        def denominator_function(z):
            return pade.B(z, self.im_axis.shape[0],
                           1j * self.im_axis, self.im_data, self.a)

        numerator = pade.A(self.re_axis, self.im_axis.shape[0],
                           1j * self.im_axis, self.im_data, self.a)
        denominator = pade.B(self.re_axis, self.im_axis.shape[0],
                           1j * self.im_axis, self.im_data, self.a)

        result = numerator / denominator

        result_dict = {}
        result_dict.update({"numerator": numerator,
                            "denominator": denominator,
                            "numerator_function": numerator_function,
                            "denominator_function": denominator_function,
                            "check": self.check(),
                            "ivcheck": self.ivcheck,
                            "A_opt": -result.imag / np.pi,
                            "g_ret": result})

        return OptimizationResult(**result_dict)


class MaxentSolverSVD(AnalyticContinuationSolver):
    """Maxent with singular-value decomposition.

    The singular-value decomposition of the kernel leads to a basis
    in which the dimensionality of the analytic continuation problem
    is much reduced. This makes computations faster and enforces some
    constraints.

    This class is never instantiated directly by the user, but instead by
    the solve method of continuation.AnalyticContinuationProblem.
    """

    def log(self, msg):
        """Print log messages if `verbose=True` was passed to the solver.

        Parameters
        ----------
        msg : str
                  Message to print
        """

        if self.verbose: print(msg)

    def __init__(self, im_axis, re_axis, im_data,
                 kernel_mode='', model=None,
                 stdev=None, cov=None,
                 offdiag=False,
                 preblur=False, blur_width=0.,
                 optimizer='newton',
                 verbose=True, **kwargs):
        """Create a Maxent solver object.

        Here we pass the data to the solver and do some precomputations.
        This makes the process of solving much faster.

        Parameters
        ----------
        im_axis : numpy.ndarray
                  One-dimensional numpy array of type float
                  Matsubara frequencies or imaginary time points of the input data.
        re_axis : numpy.ndarray
                  One-dimensional numpy array of type float
                  Real-frequency grid to use for the analytic continuation.
        im_data : numpy.ndarray
                  One-dimensional numpy array of type float or complex
                  Imaginary-axis data, e.g. Matsubara Green's function
        kernel_mode : {`'freq_fermionic'`, `'freq_bosonic'`, `'time_fermionic'`, `'time_bosonic'`}
                  * `'freq_fermionic'` fermionic Matsubara Greens function
                  * `'freq_bosonic'` bosonic Matsubara Greens function
                  * `'time_fermionic'` fermionic Green's function in imaginary time
                  * `'time_bosonic'` bosonic Green's function (susceptibility) in imaginary time

                  Additionally there are less established special-purpose options
                  [`'freq_bosonic_xyz'`, `'freq_fermionic_phsym'`, `'time_fermionic_phsym'`].
        model : numpy.ndarray
                  One-dimensional numpy array of type float.
                  (Values must be greater or equal to zero.)
                  Default model of the Maxent calculation, i.e. the entropy of the
                  spectral function is computed with respect to this model.
                  Thus the shape has to match `re_axis`.
        stdev : numpy.ndarray
                  One-dimensional numpy array of positive float values.
                  `stdev` are the standard deviation values of the measured data points,
                  thus they have to be larger than zero. The shape has to match `im_data`
                  The keywords `stdev` and `cov` are mutually exclusive, i.e. only one of them
                  can be used.
                  For complex data (e.g. Matsubara Green's function) the same standard deviation
                  is used for both the real and imaginary part, such that the array passed to this
                  keyword is taken as the standard deviation of the real part (not the absolute).
        cov : numpy.ndarray
                  Two-dimensional numpy array.
                  `cov` is the covariance matrix of the input data. If it is diagonal, the diagonal
                  elements are the variance of the data, i.e. the square of the standard deviation.
                  The keywords `stdev` and `cov` are mutually exclusive, i.e. only one of them
                  can be used.
                  For complex data, the covariance matrix can be complex (see Kappl et al., PRB 102 (8), 085124)
        offdiag : bool, default: False
                  Whether the input data are offdiagonal elements of a matrix. For diagonal elements,
                  the spectral function is positive, for offdiagonal ones it has positive and negative
                  values, and the integral is 0.
        preblur : bool, default: False
                  Whether to use preblur in the calculation
        blur_width : float, default: 0.
                  If preblur=True, specify a width of the preblur that is larger than 0.
        optimizer : {'scipy_lm', 'newton'}, default: 'newton'
                  Which optimizer to use at each value of alpha to solve the minimization / root finding problem.
                  Usually the custom implementation of the newton algorithm is much faster.
        verbose : bool, default: True
                  If set to False, no output is printed. This can be useful when doing a very large number
                  of computations, but generally I recommend leaving it on True, to see what is happening.

        """

        self.verbose = verbose

        self.kernel_mode = kernel_mode
        self.im_axis = im_axis
        self.re_axis = re_axis
        self.im_data = im_data
        self.offdiag = offdiag
        self.optimizer = optimizer

        self.nw = self.re_axis.shape[0]
        self.wmin = self.re_axis[0]
        self.wmax = self.re_axis[-1]
        self.dw = np.diff(
            np.concatenate(([self.wmin],
                            (self.re_axis[1:] + self.re_axis[:-1]) / 2.,
                            [self.wmax])))
        if not self.offdiag:
            self.model = model  # the model should be normalized by the user himself
        else:
            self.model_plus = model  # the model should be normalized by the user himself
            self.model_minus = model  # the model should be normalized by the user himself

        if cov is None and stdev is not None:
            self.var = stdev ** 2
            self.cov = np.diag(self.var)
            self.ucov = np.eye(self.im_axis.shape[0])
        elif stdev is None and cov is not None:
            self.cov = cov
            self.var, self.ucov = np.linalg.eigh(self.cov)  # go to eigenbasis of covariance matrix
            self.var = np.abs(self.var)  # numerically, var can be zero below machine precision

        self.im_data = np.dot(self.ucov.T.conj(), self.im_data)
        self.E = 1. / self.var
        self.niw = self.im_axis.shape[0]

        # set the kernel
        self.kernel = kernels.Kernel(kind=kernel_mode,
                                     im_axis=self.im_axis,
                                     re_axis=self.re_axis)

        # PREBLUR
        self.preblur = preblur
        if self.preblur:
            self.kernel.preblur(blur_width=blur_width)

        # rotate kernel to eigenbasis of covariance matrix
        self.kernel.rotate_to_cov_eb(ucov=self.ucov)

        # special treatment for complex data of fermionic frequency kernel
        if kernel_mode == 'freq_fermionic':
            self.niw *= 2
            self.im_data = np.concatenate((self.im_data.real, self.im_data.imag))
            self.var = np.concatenate((self.var, self.var))
            self.E = np.concatenate((self.E, self.E))

        # singular value decomposition of the kernel
        U, S, Vt = np.linalg.svd(self.kernel.real_matrix(), full_matrices=False)

        self.n_sv = np.arange(min(self.nw, self.niw))[S > 1e-10][-1]  # number of singular values larger than 1e-10

        self.U_svd = np.array(U[:, :self.n_sv], dtype=np.float64, order='C')
        self.V_svd = np.array(Vt[:self.n_sv, :].T, dtype=np.float64, order='C')  # numpy.svd returns V.T
        self.Xi_svd = S[:self.n_sv]

        self.log('{} data points on real axis'.format(self.nw))
        self.log('{} data points on imaginary axis'.format(self.niw))
        self.log('{} significant singular values'.format(self.n_sv))

        # =============================================================================================
        # First, precompute as much as possible
        # The precomputation of W2 is done in C, this saves a lot of time!
        # The other precomputations need less loop, can stay in python for the moment.
        # =============================================================================================
        self.log('Precomputation of coefficient matrices...')

        if not self.offdiag:  # precompute matrices W_ml (W2), W_mil (W3)
            self.W2 = np.einsum('k,km,m,kn,n,ln,l,l->ml', self.E, self.U_svd, self.Xi_svd, self.U_svd, self.Xi_svd,
                                self.V_svd, self.dw, self.model)
            self.W3 = self.W2[:, None, :] * (self.V_svd[None, :, :]).transpose((0, 2, 1))

        else:  # precompute matrices M_ml (M2), M_mil (M3)
            self.M2 = np.einsum('k,km,m,kn,n,ln,l->ml', self.E, self.U_svd, self.Xi_svd, self.U_svd, self.Xi_svd,
                                self.V_svd, self.dw)
            self.M3 = self.M2[:, None, :] * (self.V_svd[None, :, :]).transpose((0, 2, 1))

        # precompute the evidence vector Evi_m
        self.Evi = np.einsum('m,km,k,k->m', self.Xi_svd, self.U_svd, self.E, self.im_data)

        # precompute curvature of likelihood function
        self.d2chi2 = np.einsum('i,j,ki,kj,k->ij', self.dw, self.dw,
                                self.kernel.real_matrix(), self.kernel.real_matrix(),
                                self.E)

        # some arrays that are used later...
        self.chi2arr = []
        self.specarr = []
        self.backarr = []
        self.entrarr = []
        self.alpharr = []
        self.uarr = []
        self.bayesConv = []

    # =============================================================================================
    # Here, we define the main functions needed for the root finding problem
    # =============================================================================================

    def compute_f_J_diag(self, u, alpha):
        """This function evaluates the function whose root we want to find.

        The function :math:`f_m(u)` is defined as the singular value decomposition
        of the derivative :math:`dQ[A]/dA_m`. Since we want to minimize :math:`Q[A]`,
        we have to find the root of the vector-valued function :math:`f`, i.e.
        :math:`f_m(u) = SVD(dQ/dA)_m = 0`.
        For more efficient root finding, we also need the Jacobian :math:`J`.
        It is directly computed in singular space, :math:`J_{mi}=df_m/du_i`.

        Parameters
        ----------
        u : numpy.ndarray
                  singular-space vector that parametrizes the spectral function
        alpha : float
                  (positive) weight factor of the entropy

        Returns
        -------
        f : numpy.ndarray
                  value of the function whose zero we want to find
        J : numpy.ndarray
                  Jacobian at the current position
        """

        v = np.dot(self.V_svd, u)
        w = np.exp(v)
        term_1 = np.dot(self.W2, w)
        term_2 = np.dot(self.W3, w)
        f = alpha * u + term_1 - self.Evi
        J = alpha * np.eye(self.n_sv) + term_2
        return f, J

    def compute_f_J_offdiag(self, u, alpha):
        """The analogue to compute_f_J_diag for offdiagonal elements.

        Parameters
        ----------
        u : numpy.ndarray
                  singular-space vector that parametrizes the spectral function
        alpha : float
                  (positive) weight factor of the entropy

        Returns
        -------
        f : numpy.ndarray
                  value of the function whose zero we want to find
        J : numpy.ndarray
                  Jacobian at the current position
        """
        v = np.dot(self.V_svd, u)
        w = np.exp(v)
        a_plus = self.model_plus * w
        a_minus = self.model_minus / w
        a1 = a_plus - a_minus
        a2 = a_plus + a_minus
        f = alpha * u + np.dot(self.M2, a1) - self.Evi
        J = alpha * np.eye(self.n_sv) + np.dot(self.M3, a2)
        return f, J

    # =============================================================================================
    # Some auxiliary functions
    # =============================================================================================

    def singular_to_realspace_diag(self, u):
        """Go from singular to real space.

        Transform the singular space vector u into real-frequency space (spectral function)
        by :math:`A(\\omega) = D(\\omega) e^{V u}`, where :math:`D` is the default model
        and :math:`V` is the matrix from the SVD.

        Parameters
        ----------
        u : numpy.ndarray
                  singular-space vector that parametrizes the spectral function

        Returns
        -------
        numpy.ndarray
                  Spectral function :math:`A` obtained from `u`
        """
        return self.model * np.exp(np.dot(self.V_svd, u))

    def singular_to_realspace_offdiag(self, u):
        """Go from singular to real space.

        Transform the singular space vector u into real-frequency
        space in the case of an offdiagonal element.

        Parameters
        ----------
        u : numpy.ndarray
                  singular-space vector that parametrizes the spectral function

        Returns
        -------
        numpy.ndarray
                  Spectral function :math:`A` obtained from `u`
        """
        v = np.dot(self.V_svd, u)
        w = np.exp(v)
        return self.model_plus * w - self.model_minus / w

    def backtransform(self, A):
        """ Backtransformation from real to imaginary axis.

        :math:`G(i\\omega_n) = \int d\\nu \\; K(i\\omega_n, \\nu)  A(\\nu)`

        Also at this place we return from the 'diagonal-covariance space'
        Note: this function is not a bottleneck.

        Parameters
        ----------
        A : numpy.ndarray
                  Spectral function

        Returns
        -------
        np.ndarray
                  Back-transformed Green's function on imaginary axis
        """
        return np.trapz(np.dot(self.ucov, self.kernel.matrix) * A[None, :],
                        self.re_axis, axis=-1)

    def chi2(self, A):
        """compute chi-squared-deviation

        Compute the log-likelihood function or chi-squared-deviation of
        the spectral function:
        :math:`\\sum_n \\frac{|G(i\\omega_n) - \int K(i\\omega_n, \\nu) A(\\nu)|^2}{\\sigma_n^2}`

        Parameters
        ----------
        A : numpy.ndarray
                  Spectral function

        Returns
        -------
        float
                  chi-squared deviation
        """
        return np.sum(
            self.E * (self.im_data - np.trapz(self.kernel.real_matrix() * A[None, :], self.re_axis, axis=-1)) ** 2)

    def entropy_pos(self, A, u):
        """Compute entropy for positive definite spectral function.

        Parameters
        ----------
        A : numpy.ndarray
                  Spectral function
        u : numpy.ndarray
                  Singular-space vector representing the same spectral function

        Returns
        -------
        float
                  entropy"""
        return np.trapz(A - self.model - A * np.dot(self.V_svd, u), self.re_axis)

    def entropy_posneg(self, A, u):
        """Compute "positive-negative entropy" for spectral function with norm 0.

        Parameters
        ----------
        A : numpy.ndarray
                  Spectral function
        u : numpy.ndarray
                  Singular-space vector representing the same spectral function

        Returns
        -------
        float
                  entropy
        """
        root = np.sqrt(A ** 2 + 4. * self.model_plus * self.model_minus)
        return np.trapz(root - self.model_plus - self.model_minus
                        - A * np.log((root + A) / (2. * self.model_plus)),
                        self.re_axis)

    def bayes_conv(self, A, entr, alpha):
        """Bayesian convergence criterion for classic maxent (maximum of probablility distribution)

        Parameters
        ----------
        A : numpy.ndarray
                  spectral function
        entr : float
                  entropy
        alpha : float
                  weight factor of the entropy

        Returns
        -------
        ng : float
                  "number of good data points"
        tr : float
                  trace of the gamma matrix
        conv : float
                   convergence criterion (1 -> converged)
        """

        LambdaMatrix = np.sqrt(A / self.dw)[:, None] * self.d2chi2 * np.sqrt(A / self.dw)[None, :]
        try:
            lam = np.linalg.eigvalsh(LambdaMatrix)
        except np.linalg.LinAlgError:
            self.log('LinAlgError in Bayes matrix inversion. Irrelevant for chi2kink')
            lam = np.diag(LambdaMatrix)
        ng = -2. * alpha * entr
        tr = np.sum(lam / (alpha + lam))
        conv = tr / ng
        return ng, tr, conv

    def bayes_conv_offdiag(self, A, entr, alpha):
        """Bayesian convergence criterion for classic maxent, offdiagonal version

        Parameters
        ----------
        A : numpy.ndarray
                  spectral function
        entr : float
                  entropy
        alpha : float
                  weight factor of the entropy

        Returns
        -------
        ng : float
                  "number of good data points"
        tr : float
                  trace of the gamma matrix
        conv : float
                   convergence criterion (1 -> converged)
        """
        A_sq = np.power((A ** 2 + 4. * self.model_plus * self.model_minus) / self.dw ** 2, 0.25)
        LambdaMatrix = A_sq[:, None] * self.d2chi2 * A_sq[None, :]
        lam = np.linalg.eigvalsh(LambdaMatrix)
        ng = -2. * alpha * entr
        tr = np.sum(lam / (alpha + lam))
        conv = tr / ng
        return ng, tr, conv

    def posterior_probability(self, A, alpha, entr, chisq):
        """Bayesian a-posteriori probability for alpha after optimization of A

        Parameters
        ----------
        A : numpy.ndarray
                  spectral function
        entr : float
                  entropy
        alpha : float
                  weight factor of the entropy
        chisq : float
                  chi-squared deviation

        Returns
        -------
        float
                 Probability
        """
        lambda_matrix = np.sqrt(A / self.dw)[:, None] * self.d2chi2 * np.sqrt(A / self.dw)[None, :]
        try:
            lam = np.linalg.eigvalsh(lambda_matrix)
        except np.linalg.LinAlgError:
            self.log('LinAlgError in Bayes matrix inversion. Irrelevant for chi2kink')
            lam = np.diag(lambda_matrix)
        try:
            eig_sum = np.sum(np.log(alpha / (alpha + lam)))
        except RuntimeWarning:
            print(lam)
        log_prob = alpha * entr - 0.5 * chisq + np.log(alpha) + 0.5 * eig_sum
        return np.exp(log_prob)

    def maxent_optimization(self, alpha, ustart, iterfac=10000000, use_bayes=False, **kwargs):
        """optimization of maxent functional for a given value of alpha

        Since a priori the best value of :math:`\\alpha` is unknown,
        this function has to be called several times in order to find a good value.

        Parameters
        ----------
        alpha : float
                  weight factor of the entropy
        ustart : numpy.ndarray
                  singular-space vector used as a starting value for the optimization.
                  For the first optimization, done at large alpha, we use zeros,
                  which corresponds to the default model. Then we use the result of the
                  previous optimization as a starting value.
        iterfac : int, default: 10000000
                  Control parameter for maximum number of iterations in
                  scipy_lm, which is <number of singular values> * iterfac.
                  It has no effect when using the newton optimizer.
        use_bayes : bool, default=False
                  Whether to use the Bayesian inference parameters for alpha.

        Returns
        -------
        OptimizationResult
                  Object that holds the results of the optimization,
                  e.g. spectral function, chi-squared deviation.

        """

        if not self.offdiag:
            self.compute_f_J = self.compute_f_J_diag
            self.singular_to_realspace = self.singular_to_realspace_diag
            self.entropy = self.entropy_pos
        else:
            self.compute_f_J = self.compute_f_J_offdiag
            self.singular_to_realspace = self.singular_to_realspace_offdiag
            self.entropy = self.entropy_posneg

        if self.optimizer == 'newton':
            newton_solver = NewtonOptimizer(self.n_sv, initial_guess=ustart, max_hist=1)
            sol = newton_solver(self.compute_f_J, alpha)
        elif self.optimizer == 'scipy_lm':
            sol = opt.root(self.compute_f_J,  # function returning function value f and jacobian J (we search root of f)
                          ustart,  # sensible starting point
                          method='lm',  # levenberg-marquart method
                          jac=True,  # already self.compute_f_J returns the jacobian (slightly more efficient in this case)
                          options={'maxiter': iterfac * self.n_sv,  # max number of lm steps
                                   'factor': 100.,  # scale for initial stepwidth of lm (?)
                                   'diag': np.exp(np.arange(self.n_sv))},
                          # scale for values to find (assume that they decay exponentially)
                          args=(alpha))  # additional argument for self.compute_f_J
        else:
            raise ValueError("Unknown optimizer. Use either 'newton' (recommended) or 'scipy_lm'.")

        u_opt = sol.x
        A_opt = self.singular_to_realspace(sol.x)
        entr = self.entropy(A_opt, u_opt)
        chisq = self.chi2(A_opt)  # has to be applied before blurring
        norm = np.trapz(A_opt, self.re_axis)  # is not changed by blurring

        # result = OptimizationResult()
        result_dict = {}
        result_dict.update({"u_opt": u_opt})
        if self.preblur:
            result_dict.update({"A_opt": self.kernel.blur(A_opt),
                                "blur_width": self.kernel.blur_width})
            # result.A_opt = self.kernel.blur(A_opt)
            # result.blur_width = self.kernel.blur_width
        else:
            result_dict.update({"A_opt": A_opt, "blur_width": 0.})
            # result.A_opt = A_opt
            # result.blur_width = 0.
        result_dict.update({"alpha": alpha,
                            "entropy": entr,
                            "chi2": chisq,
                            "backtransform": self.backtransform(A_opt),
                            "norm": norm,
                            "Q": alpha * entr - 0.5 * chisq})
        # result.alpha = alpha
        # result.entropy = entr
        # result.chi2 = chisq
        # result.backtransform = self.backtransform(A_opt)
        # result.norm = norm
        # result.Q = alpha * entr - 0.5 * chisq

        if use_bayes:
            if not self.offdiag:
                ng, tr, conv = self.bayes_conv(A_opt, entr, alpha)
            else:
                ng, tr, conv = self.bayes_conv_offdiag(A_opt, entr, alpha)

            # result.n_good = ng
            # result.trace = tr
            # result.convergence = conv
            result_dict.update({"n_good": ng, "trace": tr, "convergence": conv})
            if not self.offdiag:
                prob = self.posterior_probability(A_opt, alpha, entr, chisq)
                # result.probability = prob
                result_dict.update({"prob": prob})

        self.log('log10(alpha) = {:3.2f},\tchi2 = {:4.3e},   S = {:4.3e},   nfev = {},   norm = {:4.3f}'.format(
            np.log10(alpha), chisq, entr, sol.nfev, norm))

        return OptimizationResult(**result_dict)

    # =============================================================================================
    # Several variants of maxent are implemented in the following
    # =============================================================================================

    def solve_historic(self):
        """ Historic Maxent: choose alpha in a way that chi^2 \approx N

        Returns
        -------
        OptimizationResult
               Result of the optimization at "best" alpha value
        list
               List of OptimizationResult objects for all used values of alpha
        """

        if not self.offdiag:
            self.compute_f_J = self.compute_f_J_diag
            self.singular_to_realspace = self.singular_to_realspace_diag
            self.entropy = self.entropy_pos
        else:
            self.compute_f_J = self.compute_f_J_offdiag
            self.singular_to_realspace = self.singular_to_realspace_offdiag
            self.entropy = self.entropy_posneg
        self.log('Solving...')
        optarr = []
        alpha = 10 ** 6
        self.ustart = np.zeros((self.n_sv))
        converged = False
        conv = 0.
        while conv < 1:
            o = self.maxent_optimization(alpha, self.ustart)
            optarr.append(o)
            alpha /= 10.
            conv = self.niw / o.chi2

        if len(optarr) <= 1:
            raise RuntimeError('Failed to get a prediction for optimal alpha. '
                               + 'Decrease the error or increase alpha_start.')

        def root_fun(alpha, u0):  # this function is just for the newton root-finding
            res = self.maxent_optimization(alpha, u0, iterfac=100000)
            optarr.append(res)
            u0[:] = res.u_opt
            return self.niw / res.chi2 - 1.

        ustart = optarr[-2].u_opt
        alpha_opt = opt.newton(root_fun, optarr[-1].alpha, tol=1e-6, args=(ustart,))
        self.log('final optimal alpha: {}, log10(alpha_opt) = '.format(alpha_opt, np.log10(alpha_opt)))

        sol = self.maxent_optimization(alpha_opt, ustart, iterfac=250000)
        self.alpha_opt = alpha_opt
        self.A_opt = sol.A_opt
        return sol, optarr


    def solve_classic(self):
        """Classic maxent alpha determination.

        Classic maxent uses Bayes statistics to approximately determine
        the most probable value of alpha
        We start at a large value of alpha, where the optimization yields basically the default model,
        therefore u_opt is only a few steps away from ustart=0 (=default model)
        Then we gradually decrease alpha, step by step moving away from the default model towards data fitting.
        Using u_opt as ustart for the next (smaller) alpha brings a great speedup into this procedure.

        Returns
        -------
        OptimizationResult
               Result of the optimization at "best" alpha value
        list
               List of OptimizationResult objects for all used values of alpha
        """
        if not self.offdiag:
            self.compute_f_J = self.compute_f_J_diag
            self.singular_to_realspace = self.singular_to_realspace_diag
            self.entropy = self.entropy_pos
        else:
            self.compute_f_J = self.compute_f_J_offdiag
            self.singular_to_realspace = self.singular_to_realspace_offdiag
            self.entropy = self.entropy_posneg
        self.log('Solving...')
        optarr = []
        alpha = 10 ** 6
        self.ustart = np.zeros((self.n_sv))
        converged = False
        conv = 0.
        while conv < 1:
            o = self.maxent_optimization(alpha, self.ustart, use_bayes=True)

            ustart = o.u_opt
            optarr.append(o)
            alpha /= 10.
            conv = o.convergence

        if len(optarr) <= 1:
            raise RuntimeError('Failed to get a prediction for optimal alpha. '
                               + 'Decrease the error or increase alpha_start.')

        bayes_conv = [o.convergence for o in optarr]
        alpharr = [o.alpha for o in optarr]

        # log(conv) is approximately a linear function from log(alpha),
        # based on this we can predict the optimal alpha quite precisely.
        exp_opt = np.log10(alpharr[-2]) \
                 - np.log10(bayes_conv[-2]) * (np.log10(alpharr[-1])
                                               - np.log10(alpharr[-2])) \
                   / (np.log10(bayes_conv[-1]) - np.log10(bayes_conv[-2]))
        alpha_opt = 10 ** exp_opt
        self.log('prediction for optimal alpha: {}, log10(alpha_opt) = {}'.format(alpha_opt, np.log10(alpha_opt)))

        # Starting from the predicted value of alpha, and starting the optimization at the solution for the next-lowest alpha,
        # we find the optimal alpha by newton's root finding method.

        def root_fun(alpha, u0):  # this function is just for the newton root-finding
            res = self.maxent_optimization(alpha, u0, iterfac=100000, use_bayes=True)
            optarr.append(res)
            u0[:] = res.u_opt
            return res.convergence - 1.

        ustart = optarr[-2].u_opt
        alpha_opt = opt.newton(root_fun, alpha_opt, tol=1e-6, args=(ustart,))
        self.log('final optimal alpha: {}, log10(alpha_opt) = '.format(alpha_opt, np.log10(alpha_opt)))

        sol = self.maxent_optimization(alpha_opt, ustart, iterfac=250000, use_bayes=True)
        self.alpha_opt = alpha_opt
        self.A_opt = sol.A_opt
        return sol, optarr


    def solve_bryan(self, alphastart=500, alphadiv=1.1, interactive=False):
        """Bryan's method of determining the optimal spectrum.

        Bryan's maxent calculates an average of spectral functions,
        weighted by their Bayesian probability

        Parameters
        ----------
        alphastart : float, default=500
                  Starting value of alpha. This is the largest value of alpha,
                  it is decreased during the calculation.
        alphadiv : float, default=1.1
                  After each optimization, the current alpha is divided by this number.
                  Hence, the number has to be larger than 1.
        interactive : bool, default=False
                  Whether to show a plot of the probability. (Needs matplotlib)

        Returns
        -------
        OptimizationResult
               Contains the weighted average of all results
        list
               List of OptimizationResult objects for all used values of alpha
        """

        if interactive:
            import matplotlib.pyplot as plt

        self.log('Solving...')
        optarr = []
        alpha = alphastart
        self.ustart = np.zeros((self.n_sv))
        maxprob = 0.
        while True:
            o = self.maxent_optimization(alpha, self.ustart, use_bayes=True)
            ustart = o.u_opt
            optarr.append(o)
            alpha /= alphadiv
            prob = o.probability
            # automatically terminate the loop when probability is small again
            if prob > maxprob:
                maxprob = prob
            elif prob < 0.01 * maxprob:
                break

        alpharr = np.array([o.alpha for o in optarr])
        probarr = np.array([o.probability for o in optarr])
        specarr = np.array([o.A_opt for o in optarr])
        probarr /= -np.trapz(probarr, alpharr)  # normalize the probability distribution
        if interactive:
            fig = plt.figure()
            plt.plot(np.log10(alpharr), probarr)
            fig.show()

        # calculate the weighted average spectrum
        A_opt = -np.trapz(specarr * probarr[:, None], alpharr,
                          axis=0)  # need a "-" sign, because alpha goes from large to small.

        sol = OptimizationResult(A_opt=A_opt,
                                 backtransform=self.backtransform(A_opt))

        return sol, optarr

    def solve_chi2kink(self, alpha_start=1e9, alpha_end=1e-3, alpha_div=10., fit_position=2.5, interactive=False, **kwargs):
        """Determine optimal alpha by searching for the kink in log(chi2) (log(alpha))

        We start with an optimization at a large value of alpha (`alpha_start`),
        where we should get only the default model. Then, alpha is decreased
        step-by-step, where :math:`\\alpha_{n+1} = \\alpha_n / \\alpha_{div}`, until the minimal
        value of alpha_end is reached.
        Then, we fit a function
        :math:`\\phi(x; a, b, c, d) = a + b / [1 + exp(-d*(x-c))]`,
        from which the optimal alpha is determined by
        x_opt = c - `fit_position` / d; alpha_opt = 10^x_opt.

        Parameters
        ----------
        alpha_start : float, default=1e9
                  Value of alpha where to start the procedure.
                  This is the largest value of alpha, it is subsequently decreased
                  in the algorithm.
        alpha_end : float, default=1e-3
                  Last (smallest) value of alpha that is considered in the algorithm
        alpha_div : float, default=10.
                  After each optimization, alpha is divided by alpha_div.
                  Thus it has to be larger than 1. The default value of 10 is
                  a good compromise of function and speed. You can take
                  smaller values if you are unsure or want to make fancy plots.
        fit_position : float, default=2.5
                  Control parameter for under/overfitting.
                  In my experience, good values are usually between 2 and 2.5.
                  Smaller values lead to underfitting, which is sometimes desirable.
                  Larger values lead to overfitting, which should be avoided.
        interactive : bool, default=False
                  Decide whether to show a plot of chi2 vs alpha.

        Returns
        -------
        OptimizationResult
               Result of the optimization at "best" alpha value
        list
               List of OptimizationResult objects for all used values of alpha
        """

        if interactive:
            import matplotlib.pyplot as plt

        alpha = alpha_start
        chi = []
        alphas = []
        optarr = []
        ustart = np.zeros((self.n_sv))
        while True:
            try:
                o = self.maxent_optimization(alpha=alpha, ustart=ustart)
                optarr.append(o)
                ustart = o.u_opt
                chi.append(o.chi2)
                alphas.append(alpha)
            except:
                # For small alphas sometimes the optimization fails
                # Usually this happens at values of alpha that
                # are too small anyway, so don't worry.
                print('Optimization at alpha={} failed.'.format(alpha))
            alpha = alpha / alpha_div
            if alpha < alpha_end:
                break

        alphas = np.asarray(alphas)
        chis = np.asarray(chi)

        def fitfun(x, a, b, c, d):
            return a + b / (1. + np.exp(-d * (x - c)))

        try:
            good_numbers = np.isfinite(chis)
            popt, pcov = opt.curve_fit(fitfun,
                                       np.log10(alphas[good_numbers]),
                                       np.log10(chis[good_numbers]),
                                       p0=(0., 5., 2., 0.))
        except ValueError:
            print('Fermi fit failed.')
            if interactive:
                plt.plot(np.log10(alphas), np.log10(chis), marker='s')
                plt.show()
                for o in optarr:
                    plt.plot(o.backtransform)
                plt.plot(self.im_data)
                plt.show()
            return optarr[-1], optarr

        a, b, c, d = popt

        if interactive:
            print('Fit parameters {}'.format(popt))
        if d < 0.:
            raise RuntimeError('Fermi fit temperature negative.')

        a_opt = c - fit_position / d
        alpha_opt = 10. ** a_opt
        self.log('Optimal log alpha {}'.format(a_opt))

        if interactive:
            plt.plot(np.log10(alphas), np.log10(chis), marker='s', label='chi2')
            plt.plot(np.log10(alphas), fitfun(np.log10(alphas), *popt), label='fit2lin')
            #plt.plot(np.log10(alphas), chi2_interp(np.log10(alphas)), label='chi2 fit')

        closest_idx = np.argmin(np.abs(np.log10(alphas) - a_opt))
        ustart = optarr[closest_idx].u_opt
        sol = self.maxent_optimization(alpha_opt, ustart)

        if interactive:
            plt.plot(a_opt, np.log10(sol.chi2), marker='s', color='red', label='opt')
            plt.legend()
            plt.show()

        return sol, optarr

    def error_propagation(self, obs, args):
        """Calculate the deviation for a callable function obs.
        This feature is still experimental"""
        hess = -self.d2chi2
        hess += self.alpha_opt * np.diag(1. / self.A_opt) \
                * self.dw[:, None] * self.dw[None, :]  # possibly, it's 2*alpha
        varmat = np.linalg.inv(hess)  # TODO actually here should be a minus...
        obsval = obs(self.re_axis, args)
        integrand = obsval[:, None] * obsval[None, :] \
                    * varmat * self.dw[:, None] * self.dw[None, :]
        deviation = np.sum(integrand)
        return deviation

    def solve(self, **kwargs):
        """Wrapper function for solve, which calls the chosen
        method of alpha_determination."""
        try:
            alpha_determination = kwargs['alpha_determination']
        except KeyError:
            raise KeyError("No valid alpha determination mode found. Recommended: alpha_determination='chi2kink'")
        if alpha_determination == 'historic':
            return self.solve_historic()
        elif alpha_determination == 'classic':
            return self.solve_classic()
        elif alpha_determination == 'bryan':
            return self.solve_bryan()
        elif alpha_determination == 'chi2kink':
            return self.solve_chi2kink(**kwargs)
        else:
            raise ValueError('Unknown alpha determination mode')


class OptimizationResult(object):
    """Object for holding the result of an optimization.

    This class has no methods except the constructor,
    it is thus essentially a collection of output numbers.

    All member variables have None as default value, different solvers
    override different variables. A_opt is always set, since it
    is the main result of analytic continuation.
    """

    def __init__(self,
                 u_opt=None,
                 A_opt=None,
                 chi2=None,
                 backtransform=None,
                 entropy=None,
                 n_good=None,
                 probability=None,
                 alpha=None,
                 convergence=None,
                 trace=None,
                 Q=None,
                 norm=None,
                 blur_width=None,
                 numerator=None,
                 denominator=None,
                 numerator_function=None,
                 denominator_function=None,
                 check=None,
                 ivcheck=None,
                 g_ret=None):
        self.u_opt = u_opt
        self.A_opt = A_opt
        self.chi2 = chi2
        self.backtransform = backtransform
        self.entropy = entropy
        self.n_good = n_good
        self.probability = probability
        self.alpha = alpha
        self.convergence = convergence
        self.trace = trace
        self.Q = Q
        self.norm = norm
        self.blur_width = blur_width
        self.numerator=numerator
        self.denominator=denominator
        self.numerator_function=numerator_function
        self.denominator_function=denominator_function
        self.check = check
        self.ivcheck = ivcheck
        self.g_ret = g_ret


class NewtonOptimizer(object):
    """Newton root finding."""

    def __init__(self, opt_size, max_hist=1, max_iter=20000, initial_guess=None):
        """
        :param opt_size: number of variables (integer)
        :param max_hist: maximal history for mixing (integer)
        :param max_iter: maximum number of iterations for root finding
        :param initial_guess: initial guess for the root finding.
        """

        if initial_guess is None:
            initial_guess = np.zeros((opt_size))

        self.props = [initial_guess]
        self.res = []
        self.max_hist = max_hist
        self.max_iter = max_iter
        self.opt_size = opt_size
        self.return_object = collections.namedtuple("NewtonResult", ['x', 'nfev'])

    def iteration_function(self, proposal, function_vector, jacobian_matrix):
        """The function, whose fixed point we are searching.

        This function generates the iteration procedure in the Newton root finding.
        Basically, it computes the "result" from the "proposal".
        It has the form result = proposal + increment, but
        since the increment may be large, we apply a reduction of step width in such cases.
        """

        try:
            increment = -np.dot(np.linalg.pinv(jacobian_matrix), function_vector)
        except np.linalg.LinAlgError:
            print('LinAlgError in Newton Solver, setting increment to zero')
            increment = np.zeros_like(proposal)
        step_reduction = 1.
        significance_limit = 1e-4
        if np.any(np.abs(proposal) > significance_limit):
            ratio = np.abs(increment / proposal)
            max_ratio = np.amax(ratio[np.abs(proposal) > significance_limit])
            if max_ratio > 1.:
                step_reduction = 1. / max_ratio
        result = proposal + step_reduction * increment
        return result

    def __call__(self, function_and_jacobian, alpha):
        """Main function of Newton optimization.

        This function implements the self-consistent iteration of the root finding.

        Returns
        -------
        collections.namedtuple
               sol.x is the result
               sol.nfev is the number of function evaluations
        """
        f, J = function_and_jacobian(self.props[0], alpha)
        initial_result = self.iteration_function(self.props[0], f, J)
        self.res.append(initial_result)

        counter = 0
        converged = False
        while not converged:
            prop = self.get_proposal()
            f, J = function_and_jacobian(prop, alpha)
            result = self.iteration_function(prop, f, J)
            self.props.append(prop)
            self.res.append(result)
            converged = (counter > self.max_iter or np.max(np.abs(result - prop)) < 1e-4)
            counter += 1
            if np.any(np.isnan(result)):
                raise RuntimeWarning('Function returned NaN.')
        if counter > self.max_iter:
            raise RuntimeWarning('Failed to get optimization result in {} iterations'.format(self.max_iter))

        self.return_object.x = result
        self.return_object.nfev = counter
        return self.return_object

    def get_proposal(self, mixing=0.5):
        """Propose a new solution by linear mixing."""

        n_iter = len(self.props)
        history = min(self.max_hist, n_iter) - 1

        new_proposal = self.props[n_iter - 1]
        f_i = self.res[n_iter - 1] - self.props[n_iter - 1]
        update = mixing * f_i
        return new_proposal + update



from Cython.Distutils import build_ext
from distutils.core import setup, Extension
import numpy
from Cython.Build import cythonize

cythonize("ana_cont/pade.pyx")
setup(name='ana_cont',
      version='1.0',
      description='Analytic continuation package',
      author='Josef Kaufmann',
      author_email='josefkaufma@gmail.com',
      url='https://github.com/josefkaufmann/ana_cont',
      packages=['ana_cont', 'gui'],
      package_dir={'ana_cont': 'ana_cont/', 'gui': 'gui/'},
      cmdclass={'build_ext': build_ext},
      ext_modules=[Extension('ana_cont.pade',
                             sources=['ana_cont/pade.c'],
                             libraries=['m'],
                             include_dirs=[numpy.get_include()])],
      include_package_data=True,
      zip_safe=False,
      setup_requires=['Cython'],
      install_requires=['numpy', 'scipy', 'Cython', 'h5py', 'matplotlib', 'PyQt5'],
      scripts=['scripts/maxent.py', 'scripts/pade.py', 'scripts/maxent_bosonic.py']
      )




#!/usr/bin/env python
import sys, os, traceback
from PyQt5 import QtWidgets

sys.excepthook = traceback.print_exception

file_dir = os.path.dirname(os.path.abspath(__file__))
package_dir = '/'.join(file_dir.split('/')[:-1])
sys.path.insert(0, package_dir)
from gui.maxent_bosonic_frontend import MainWindow

if __name__ == "__main__":
    app = QtWidgets.QApplication(sys.argv)
    window = MainWindow()
    window.show()
    sys.exit(app.exec_())